# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import lightning.pytorch as pl
from lightning.pytorch.utilities.exceptions import MisconfigurationException
from lightning.pytorch.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from torch.utils import data

from nemo.lightning.data import WrappedDataLoader
from nemo.lightning.io.mixin import IOMixin
from nemo.lightning.pytorch.plugins import MegatronDataSampler

if TYPE_CHECKING:
    from megatron.core.datasets.bert_dataset import BERTMaskedWordPieceDatasetConfig

    from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec


class BERTPreTrainingDataModule(pl.LightningDataModule, IOMixin):
    """PyTorch Lightning-compatible data module for pre-training
       BERT-style models.
    Args:
        paths (Path | List | Dict[str, List]): Paths of the data distributions. Can be either a
            single path, a list of paths, or a dictionary. If a single path or a list of paths,
            the given paths will be used to generate the train, validation and test datasets. If
            providing a list of paths, the format can be either (1) a list of paths, e.g.
                ["path/to/dataset_1_prefix", "path/to/dataset_2_prefix"],
            or (2) a flattened, zipped list of weights and paths, e.g.
                ["30", "path/to/dataset_1_prefix", "70", "path/to/dataset_2_prefix"]
            If a dictionary is provided, it is expected to have the following form:
                {
                    'train': <TRAIN PATHS>,
                    'validation': <VALID PATHS>,
                    'test': <TEST PATHS>
                }
            where each value is either a path or a list of paths as described above.
            In this case, each split will be generated using the given paths.
            Note that if limit_val_batches <= 1, we generate the entire validaton dataset, so
            weights should not be provided for the validation split.
        seq_length (int): Sequence length.
        tokenizer (Optional["TokenizerSpec"]): An instance of a TokenizerSpec object.
        micro_batch_size (int): Batch size per GPU.
        global_batch_size (int): Global batch size.
        rampup_batch_size (Optional[List[int]]): Rampup batch size, should be in format of
            [start_global_batch_size, batch_size_increment, ramup_samples].
        num_workers (int): See ``torch.utils.data.DataLoader`` documentation.
        pin_memory (bool): See ``torch.utils.data.DataLoader`` documentation.
        persistent_workers (bool): See ``torch.utils.data.DataLoader`` documentation.
        reset_position_ids (bool): Option to reset the position IDs in the dataset at an interval.
        reset_attention_mask (bool): Option to reset the attention mask from the dataset.
        eod_mask_loss (int): Option to enable the EOD mask loss.
        seed (int): Seed for generating the GPT dataset.
        split (str): A string of 3 comma-separated integers denoting how much of the distribution
            to allocate to train, validation, and test sets, respectively. Unused if ``paths`` is a dict.
        index_mapping_dir (Optional[str]): Path to a directory to write index mapping files.
    """

    def __init__(
        self,
        paths: Union[Path, List, Dict[str, List]],
        seq_length: int = 2048,
        tokenizer: Optional["TokenizerSpec"] = None,
        micro_batch_size: int = 4,
        global_batch_size: int = 8,
        rampup_batch_size: Optional[List[int]] = None,
        num_workers: int = 8,
        pin_memory: bool = True,
        persistent_workers: bool = False,
        reset_position_ids: bool = False,
        reset_attention_mask: bool = False,
        eod_mask_loss: bool = False,
        seed: int = 1234,
        split: str = "900,50,50",
        index_mapping_dir: Optional[str] = None,
    ) -> None:
        super().__init__()
        if not isinstance(paths, (list, tuple, dict)):
            paths = [paths]

        from megatron.core.datasets.utils import get_blend_from_list

        build_kwargs = {}
        if isinstance(paths, dict):
            if split is not None:
                warnings.warn(
                    f"{split=} will be ignored since datasets are being created " f"from 3 separate distributions."
                )
            build_kwargs["blend_per_split"] = [
                get_blend_from_list(paths["train"]),
                get_blend_from_list(paths["validation"]),
                get_blend_from_list(paths["test"]),
            ]
        else:
            paths, weights = get_blend_from_list(paths)
            if len(paths) == 1:
                weights = None
            build_kwargs["blend"] = [paths, weights]
            build_kwargs["split"] = split

        self.build_kwargs = build_kwargs
        self.seq_length = seq_length
        self.micro_batch_size = micro_batch_size
        self.global_batch_size = global_batch_size
        self.tokenizer = tokenizer
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers
        self.reset_position_ids = reset_position_ids
        self.reset_attention_mask = reset_attention_mask
        self.eod_mask_loss = eod_mask_loss
        self.seed = seed
        self.split = split
        self.index_mapping_dir = index_mapping_dir
        self.init_global_step = 0

        from nemo.collections.common.tokenizers.tokenizer_utils import get_nmt_tokenizer

        self.tokenizer = tokenizer or get_nmt_tokenizer("megatron", "BertWordPieceLowerCase")

        self.data_sampler = MegatronDataSampler(
            seq_len=self.seq_length,
            micro_batch_size=self.micro_batch_size,
            global_batch_size=self.global_batch_size,
            rampup_batch_size=rampup_batch_size,
        )

    def setup(self, stage: str = "") -> None:
        """Assign Train/Val/Test dataset"""
        from megatron.core.datasets.bert_dataset import BERTMaskedWordPieceDataset
        from megatron.core.datasets.blended_megatron_dataset_builder import BlendedMegatronDatasetBuilder

        assert (
            hasattr(self, "trainer") and self.trainer is not None
        ), "Setup should be completed when trainer and config are attached."

        # Trainer API
        max_train_steps = self.trainer.max_steps
        assert max_train_steps > 0, "Please specify trainer.max_steps"
        eval_iters = (max_train_steps // self.trainer.val_check_interval + 1) * self.trainer.limit_val_batches
        test_iters = self.trainer.limit_test_batches
        num_train_samples = int(max_train_steps * self.data_sampler.global_batch_size)
        num_val_samples = int(eval_iters * self.data_sampler.global_batch_size)
        num_test_samples = int(test_iters * self.data_sampler.global_batch_size)

        if self.trainer.limit_val_batches <= 1.0 and isinstance(self.trainer.limit_val_batches, float):
            assert "blend" not in self.build_kwargs, (
                "When using a single data distribution, limit_val_batches <= 1.0 is not supported. If you'd "
                "like to run with a fractional value of limit_val_batches, please pass in separate datasets for "
                "the train, validation, and test datasets by providing a dictionary of paths, e.g.: \n"
                "    paths={ \n "
                "        'train': [PATHS FOR TRAIN], \n "
                "        'validation': [PATHS FOR VALIDATION], \n "
                "        'test' :[PATHS FOR TEST],  \n"
                "    }"
            )

            # This is to make sure we only have one epoch on every validation iteration
            num_val_samples = None

        train_valid_test_num_samples = [num_train_samples, num_val_samples, num_test_samples]
        self._train_ds, self._validation_ds, self._test_ds = BlendedMegatronDatasetBuilder(
            BERTMaskedWordPieceDataset,
            train_valid_test_num_samples,
            is_built_on_rank=lambda: True,
            config=self.bert_dataset_config,
        ).build()

    # uncomment once fabric API is merged
    # def fabric_setup(
    #     self,
    #     fabric: fl.Fabric,
    #     num_train_samples: int,
    #     num_val_samples: int,
    #     num_test_samples: int,
    # ) -> None:
    #     from megatron.core.datasets.blended_megatron_dataset_builder import BlendedMegatronDatasetBuilder
    #     from megatron.core.datasets.gpt_dataset import GPTDataset
    #
    #     del fabric
    #     train_valid_test_num_samples = [num_train_samples, num_val_samples, num_test_samples]
    #     self._train_ds, self._validation_ds, self._test_ds = BlendedMegatronDatasetBuilder(
    #         GPTDataset, train_valid_test_num_samples, self.gpt_dataset_config,
    #     ).build()

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """Create Train dataloader"""
        return self._create_dataloader(self._train_ds, mode='train')

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """Create Validation dataloader"""
        return self._create_dataloader(self._validation_ds, mode='validation')

    def test_dataloader(self) -> EVAL_DATALOADERS:
        """Create Test dataloader"""
        return self._create_dataloader(self._test_ds, mode='test')

    def _create_dataloader(self, dataset, mode, **kwargs) -> WrappedDataLoader:
        self.init_global_step = self.trainer.global_step
        self.data_sampler.init_global_step = self.init_global_step
        dataloader = WrappedDataLoader(
            mode=mode,
            dataset=dataset,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
            collate_fn=getattr(dataset, 'collate_fn', data.dataloader.default_collate),
            **kwargs,
        )
        return dataloader

    @property
    def bert_dataset_config(self) -> "BERTMaskedWordPieceDatasetConfig":
        """Create Bert Dataset Config using Mcore's BERT MaskedWordPieceDatasetConfig"""
        from megatron.core.datasets.bert_dataset import BERTMaskedWordPieceDatasetConfig

        return BERTMaskedWordPieceDatasetConfig(
            random_seed=self.seed,
            sequence_length=self.seq_length,
            tokenizer=self.tokenizer,
            path_to_cache=self.index_mapping_dir,
            classification_head=True,
            masking_probability=0.15,
            short_sequence_probability=0.10,
            masking_max_ngram=3,  # Following values are taken from megatron-lm/pretrain_bert.py
            masking_do_full_word=True,
            masking_do_permutation=False,
            masking_use_longer_ngrams=False,
            masking_use_geometric_distribution=False,
            **self.build_kwargs,
        )

    def state_dict(self) -> Dict[str, Any]:
        """Called when saving a checkpoint, implement to generate and save datamodule state.

        Returns:
            A dictionary containing datamodule state.

        """
        consumed_samples = self.data_sampler.compute_consumed_samples(self.trainer.global_step - self.init_global_step)
        return {'consumed_samples': consumed_samples}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Called when loading a checkpoint, implement to reload datamodule state given datamodule stat

        Args:
            state_dict: the datamodule state returned by ``state_dict``.

        """
        try:
            from megatron.core.num_microbatches_calculator import update_num_microbatches

        except (ImportError, ModuleNotFoundError):
            logging.warning("Megatron num_microbatches_calculator not found, using Apex version.")
            from apex.transformer.pipeline_parallel.utils import update_num_microbatches

        consumed_samples = state_dict['consumed_samples']
        self.data_sampler.init_consumed_samples = consumed_samples
        self.data_sampler.prev_consumed_samples = consumed_samples

        update_num_microbatches(
            consumed_samples=consumed_samples,
            consistency_check=False,
        )
        self.data_sampler.if_first_step = 1

    def reconfigure_limit_batches(self):
        """Reconfigure trainer.limit_val_batches for pretraining"""
        # Override limit_train_batches in terms of num of microbatches
        self._reconfigure_limit_batches(self.trainer.limit_train_batches, self._train_ds, 'train')
        # Override limit_val_batches to be a multiple of num microbatches
        # to prevent val_step from exiting in between a step
        self._reconfigure_limit_batches(self.trainer.limit_val_batches, self._validation_ds, 'val')

    def _reconfigure_limit_batches(self, limit_batches, dataloader, mode):
        """
        Reconfigure trainer.limit_val_batches for pretraining
        """
        # Override limit_batches in terms of num microbatches
        # and so there are limit_batches//num_micro_batches num of global batches
        try:
            from megatron.core.num_microbatches_calculator import get_num_microbatches

        except (ImportError, ModuleNotFoundError):
            logging.warning("Megatron num_microbatches_calculator not found, using Apex version.")
            from apex.transformer.pipeline_parallel.utils import get_num_microbatches

        if isinstance(limit_batches, int):
            limit_batches *= get_num_microbatches()
        else:
            assert isinstance(limit_batches, float)
            # Don't reconfigure if limit_batches is 0.0 or if there's no dataloader
            if limit_batches == 0.0 or dataloader is None:
                return
            # len(dataloader) returns len as num of microbatches
            dl_len_in_micro_batches = len(dataloader)
            if len(dataloader) != float("inf"):
                if limit_batches == 1.0:
                    limit_batches = dl_len_in_micro_batches
                else:
                    limit_micro_batches = int(dl_len_in_micro_batches * limit_batches)
                    if limit_micro_batches == 0 and limit_batches > 0.0:
                        min_percentage = 1.0 / len(dataloader)
                        raise MisconfigurationException(
                            f"You requested to check {limit_batches} of the val_dataloader but"
                            f" {limit_batches} * {len(dataloader)} < 1. Please increase the"
                            f" `limit_val_batches` argument. Try at least"
                            f" `limit_val_batches={min_percentage}`"
                        )
                    # Make sure trainer.limit_val_batches is a multiple of num of microbatches
                    if limit_micro_batches < get_num_microbatches():
                        limit_batches = get_num_microbatches()
                    else:
                        limit_batches = limit_batches - limit_batches % get_num_microbatches()

        if mode == 'train':
            self.trainer.limit_train_batches = limit_batches
        else:
            self.trainer.limit_val_batches = limit_batches

        # Override num sanity steps to be a multiple of num of microbatches
        self.trainer.num_sanity_val_steps *= get_num_microbatches()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import einops
import pytest
import torch

from nemo.collections.audio.parts.submodules.conformer import SpectrogramConformer
from nemo.collections.audio.parts.submodules.ncsnpp import (
    NoiseConditionalScoreNetworkPlusPlus,
    SpectrogramNoiseConditionalScoreNetworkPlusPlus,
)
from nemo.collections.audio.parts.submodules.transformerunet import SpectrogramTransformerUNet, TransformerUNet


@pytest.fixture(params=[True, False], ids=["conditioned_on_time", "not_conditioned_on_time"])
def ncsnpp(request):
    return NoiseConditionalScoreNetworkPlusPlus(
        in_channels=2, out_channels=1, num_resolutions=2, channels=(16, 16, 16), conditioned_on_time=request.param
    )


@pytest.fixture(params=[True, False], ids=["conditioned_on_time", "not_conditioned_on_time"])
def transformerunet(request):
    dim = 16
    return TransformerUNet(
        dim=dim, depth=2, heads=4, ff_mult=2, adaptive_rmsnorm=request.param, adaptive_rmsnorm_cond_dim_in=dim
    )


@pytest.fixture(params=[True, False], ids=["conditioned_on_time", "not_conditioned_on_time"])
def spectrogram_ncsnpp(request):
    return SpectrogramNoiseConditionalScoreNetworkPlusPlus(
        in_channels=2, out_channels=1, num_resolutions=2, channels=(16, 16, 16), conditioned_on_time=request.param
    )


@pytest.fixture(params=[True, False], ids=["conditioned_on_time", "not_conditioned_on_time"])
def spectrogram_transformerunet(request):
    return SpectrogramTransformerUNet(
        in_channels=2, out_channels=1, freq_dim=16, dim=16, depth=2, heads=4, ff_mult=2, adaptive_rmsnorm=request.param
    )


@pytest.fixture()
def spectrogram_conformer():
    return SpectrogramConformer(
        in_channels=2,
        out_channels=1,
        feat_in=16,
        feat_out=16,
        n_layers=2,
        d_model=16,
        conv_kernel_size=3,
        subsampling_factor=1,
    )


@pytest.fixture()
def mock_input_3d():
    batch_size = 3
    input_dim = 16
    time_steps = 20
    with torch.random.fork_rng():
        torch.random.manual_seed(0)
        input_ = torch.randn(batch_size, input_dim, time_steps)
        input_length = torch.randint(low=1, high=time_steps, size=(batch_size,))
        condition = torch.ones(batch_size).float()
    return input_, input_length, condition


@pytest.fixture()
def mock_input_4d():
    batch_size = 3
    channels = 2
    input_dim = 16
    time_steps = 20
    with torch.random.fork_rng():
        torch.random.manual_seed(0)
        input_ = torch.randn(batch_size, channels, input_dim, time_steps)
        input_length = torch.randint(low=1, high=time_steps, size=(batch_size,))
        condition = torch.ones(batch_size).float()
    return input_, input_length, condition


def test_ncsnpp_forward(ncsnpp, mock_input_4d):
    input_, input_length, condition = mock_input_4d
    batch_size, _, input_dim, time_steps = input_.shape

    output, output_length = ncsnpp(
        input=input_, input_length=input_length, condition=condition if ncsnpp.conditioned_on_time else None
    )
    assert output.shape[0] == batch_size
    assert output.shape[1] == ncsnpp.out_channels
    assert output.shape[2] == input_dim
    assert output.shape[3] == time_steps
    assert torch.all(output_length == input_length)


def test_transformerunet_forward(transformerunet, mock_input_3d):
    input_, _, _ = mock_input_3d
    input_ = einops.rearrange(input_, "B D T -> B T D")

    batch_size, *_ = input_.shape

    if transformerunet.adaptive_rmsnorm:
        adaptive_rmsnorm_cond = torch.ones((batch_size, transformerunet.adaptive_rmsnorm_cond_dim_in)).float()
    else:
        adaptive_rmsnorm_cond = None

    output = transformerunet(x=input_, adaptive_rmsnorm_cond=adaptive_rmsnorm_cond)
    assert output.shape == input_.shape


def test_spectrogram_ncsnpp_forward(spectrogram_ncsnpp, mock_input_4d):
    input_, input_length, condition = mock_input_4d
    input_ = torch.view_as_complex(torch.stack([input_, input_], dim=-1)).contiguous()

    batch_size, _, input_dim, time_steps = input_.shape

    output, output_length = spectrogram_ncsnpp(
        input=input_,
        input_length=input_length,
        condition=condition if spectrogram_ncsnpp.ncsnpp.conditioned_on_time else None,
    )
    assert output.shape[0] == batch_size
    assert output.shape[1] == spectrogram_ncsnpp.out_channels
    assert output.shape[2] == input_dim
    assert output.shape[3] == time_steps
    assert torch.all(output_length == input_length)


def test_spectrogram_transformerunet_forward(spectrogram_transformerunet, mock_input_4d):
    input_, input_length, condition = mock_input_4d
    input_ = torch.view_as_complex(torch.stack([input_, input_], dim=-1)).contiguous()

    batch_size, _, input_dim, time_steps = input_.shape

    output, output_length = spectrogram_transformerunet(
        input=input_,
        input_length=input_length,
        condition=condition if hasattr(spectrogram_transformerunet, 'sinu_pos_emb') else None,
    )
    assert output.shape[0] == batch_size
    assert output.shape[1] == spectrogram_transformerunet.out_channels
    assert output.shape[2] == input_dim
    assert output.shape[3] == time_steps
    assert torch.all(output_length == input_length)


def test_spectrogram_conformer_forward(spectrogram_conformer, mock_input_4d):
    input_, input_length, condition = mock_input_4d
    input_ = torch.view_as_complex(torch.stack([input_, input_], dim=-1)).contiguous()

    batch_size, _, input_dim, time_steps = input_.shape

    output, output_length = spectrogram_conformer(input=input_, input_length=input_length)
    assert output.shape[0] == batch_size
    assert output.shape[1] == spectrogram_conformer.out_channels
    assert output.shape[2] == input_dim
    assert output.shape[3] == time_steps
    assert torch.all(output_length == input_length)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch
import torch.nn.functional as F
from torch import nn

from nemo.collections.tts.modules.attribute_prediction_model import get_attribute_prediction_model
from nemo.collections.tts.modules.common import (
    AffineTransformationLayer,
    BiLSTM,
    ConvAttention,
    ExponentialClass,
    Invertible1x1Conv,
    Invertible1x1ConvLUS,
    LinearNorm,
    get_radtts_encoder,
)
from nemo.collections.tts.parts.utils.helpers import get_mask_from_lengths, mas_width1, regulate_len
from nemo.core.classes import Exportable, NeuralModule
from nemo.core.neural_types.elements import Index, LengthsType, MelSpectrogramType, TokenDurationType, TokenIndex
from nemo.core.neural_types.neural_type import NeuralType


@torch.jit.script
def pad_dur(dur, txt_enc):
    if dur.shape[-1] < txt_enc.shape[-1]:
        to_pad = txt_enc.shape[-1] - dur.shape[-1]
        dur = F.pad(dur, [0, to_pad])
    return dur


@torch.jit.script
def pad_energy_avg_and_f0(energy_avg, f0, max_out_len):
    to_pad = int(max_out_len - energy_avg.shape[1])
    if to_pad > 0:
        f0 = F.pad(f0[None], [0, to_pad])[0]
        energy_avg = F.pad(energy_avg[None], [0, to_pad])[0]
    to_pad = int(max_out_len - f0.shape[1])
    if to_pad > 0:
        f0 = F.pad(f0[None], [0, to_pad])[0]
    return energy_avg, f0


def adjust_f0(f0, f0_mean, f0_std, vmask_bool, musical_scaling=True):
    if f0_mean > 0.0:
        if musical_scaling:
            f0_mu, f0_sigma = f0[vmask_bool].mean(), f0[vmask_bool].std()
            f0_factor = f0_mean / f0_mu
            f0[vmask_bool] *= f0_factor
        else:
            f0_sigma, f0_mu = torch.std_mean(f0[vmask_bool])
            f0 = ((f0 - f0_mu) / f0_sigma).to(dtype=f0.dtype)
            f0_std = f0_std if f0_std > 0 else f0_sigma
            f0 = (f0 * f0_std + f0_mean).to(dtype=f0.dtype)
            f0 = f0.masked_fill(~vmask_bool, 0.0)
    return f0


class FlowStep(nn.Module):
    def __init__(
        self,
        n_mel_channels,
        n_context_dim,
        n_layers,
        affine_model='simple_conv',
        scaling_fn='exp',
        matrix_decomposition='',
        affine_activation='softplus',
        use_partial_padding=False,
    ):
        super(FlowStep, self).__init__()
        if matrix_decomposition == 'LUS':
            self.invtbl_conv = Invertible1x1ConvLUS(n_mel_channels)
        else:
            self.invtbl_conv = Invertible1x1Conv(n_mel_channels)

        self.affine_tfn = AffineTransformationLayer(
            n_mel_channels,
            n_context_dim,
            n_layers,
            affine_model=affine_model,
            scaling_fn=scaling_fn,
            affine_activation=affine_activation,
            use_partial_padding=use_partial_padding,
        )

    def forward(self, z, context, inverse=False, seq_lens=None):
        if inverse:  # for inference z-> mel
            z = self.affine_tfn(z, context, inverse, seq_lens=seq_lens)
            z = self.invtbl_conv(z, inverse)
            return z
        else:  # training mel->z
            z, log_det_W = self.invtbl_conv(z)
            z, log_s = self.affine_tfn(z, context, seq_lens=seq_lens)
            return z, log_det_W, log_s


class RadTTSModule(NeuralModule, Exportable):
    """
    Takes model parameters (modelConfig) from config file to initialize radtts module.
    Specify the type of training in the include_modules parameter. "decatnvpred" for decoder training. and "decatnunvbiasdpmvpredapm" for feature training
    n_speakers (int): Number of speakers
    n_speaker_dim (int): number of speakers dimension
    n_text (int): Symbols embedding size
    n_text_dim (int):
    n_flows (int):
    n_conv_layers_per_step (int): number of convolution layers per step
    dummy_speaker_embedding (bool):
    include_modules (string): A string that describes what to train. "decatnvpred" for decoder training. and "decatnunvbiasdpmvpredapm" for feature training.
    scaling_fn (string): scaling function
    decoder_use_partial_padding (Bool): Set this to True to add partial padding
    learn_alignments (Bool): set this to true to learn alignments
    attn_use_CTC (Bool): set True to use CTC
    n_f0_dims (int): number of Pitch dimension
    n_early_size (int):
    n_early_every (int):
    n_group_size (int):
    decoder_use_unvoiced_bias (bool):
    context_lstm_w_f0_and_energy (bool):
    use_first_order_features (bool):
    ap_pred_log_f0 (bool):
    dur_model_config: model configuration for duration
    f0_model_config: model configuration for Pitch
    energy_model_config: model configuration for energy
    """

    def __init__(
        self,
        n_speakers,
        n_speaker_dim,
        n_text,
        n_text_dim,
        n_flows,
        n_conv_layers_per_step,
        n_mel_channels,
        dummy_speaker_embedding,
        n_early_size,
        n_early_every,
        n_group_size,
        affine_model,
        dur_model_config,
        f0_model_config,
        energy_model_config,
        v_model_config=None,
        include_modules='dec',
        scaling_fn='exp',
        matrix_decomposition='',
        learn_alignments=False,
        affine_activation='softplus',
        attn_use_CTC=True,
        use_context_lstm=False,
        context_lstm_norm=None,
        n_f0_dims=0,
        n_energy_avg_dims=0,
        context_lstm_w_f0_and_energy=True,
        use_first_order_features=False,
        unvoiced_bias_activation='',
        ap_pred_log_f0=False,
        **kwargs,
    ):
        super(RadTTSModule, self).__init__()
        assert n_early_size % 2 == 0
        self.n_mel_channels = n_mel_channels
        self.n_f0_dims = n_f0_dims  # >= 1 to trains with f0
        self.n_energy_avg_dims = n_energy_avg_dims  # >= 1 trains with energy
        self.decoder_use_partial_padding = kwargs['decoder_use_partial_padding']
        self.n_speaker_dim = n_speaker_dim
        assert self.n_speaker_dim % 2 == 0
        self.speaker_embedding = torch.nn.Embedding(n_speakers, self.n_speaker_dim)
        self.embedding = torch.nn.Embedding(n_text, n_text_dim)
        self.flows = torch.nn.ModuleList()
        self.encoder = get_radtts_encoder(encoder_embedding_dim=n_text_dim)
        self.dummy_speaker_embedding = dummy_speaker_embedding
        self.learn_alignments = learn_alignments
        self.affine_activation = affine_activation
        self.include_modules = include_modules
        self.attn_use_CTC = bool(attn_use_CTC)
        self.use_context_lstm = bool(use_context_lstm)
        self.context_lstm_norm = context_lstm_norm
        self.context_lstm_w_f0_and_energy = context_lstm_w_f0_and_energy
        self.use_first_order_features = bool(use_first_order_features)
        self.decoder_use_unvoiced_bias = kwargs['decoder_use_unvoiced_bias']
        self.ap_pred_log_f0 = ap_pred_log_f0
        self.ap_use_unvoiced_bias = kwargs['ap_use_unvoiced_bias']

        if 'atn' in include_modules or 'dec' in include_modules:
            if self.learn_alignments:
                self.attention = ConvAttention(n_mel_channels, self.n_speaker_dim, n_text_dim)

            self.n_flows = n_flows
            self.n_group_size = n_group_size

            n_flowstep_cond_dims = self.n_speaker_dim + (n_text_dim + n_f0_dims + n_energy_avg_dims) * n_group_size

            if self.use_context_lstm:
                n_in_context_lstm = self.n_speaker_dim + n_text_dim * n_group_size
                n_context_lstm_hidden = int((self.n_speaker_dim + n_text_dim * n_group_size) / 2)

                if self.context_lstm_w_f0_and_energy:
                    n_in_context_lstm = n_f0_dims + n_energy_avg_dims + n_text_dim
                    n_in_context_lstm *= n_group_size
                    n_in_context_lstm += self.n_speaker_dim
                    n_flowstep_cond_dims = self.n_speaker_dim + n_text_dim * n_group_size

                self.context_lstm = BiLSTM(
                    input_size=n_in_context_lstm, hidden_size=n_context_lstm_hidden, num_layers=1,
                )

            if self.n_group_size > 1:
                self.unfold_params = {
                    'kernel_size': (n_group_size, 1),
                    'stride': n_group_size,
                    'padding': 0,
                    'dilation': 1,
                }
                self.unfold_mod = nn.Unfold(**self.unfold_params)

            self.exit_steps = []
            self.n_early_size = n_early_size
            n_mel_channels = n_mel_channels * n_group_size

            for i in range(self.n_flows):
                if i > 0 and i % n_early_every == 0:  # early exiting
                    n_mel_channels -= self.n_early_size
                    self.exit_steps.append(i)

                self.flows.append(
                    FlowStep(
                        n_mel_channels,
                        n_flowstep_cond_dims,
                        n_conv_layers_per_step,
                        affine_model,
                        scaling_fn,
                        matrix_decomposition,
                        affine_activation=affine_activation,
                        use_partial_padding=self.decoder_use_partial_padding,
                    )
                )

        if 'dpm' in include_modules:
            dur_model_config['hparams']['n_speaker_dim'] = n_speaker_dim
            self.dur_pred_layer = get_attribute_prediction_model(dur_model_config)

        self.use_unvoiced_bias = False
        self.use_vpred_module = False
        self.ap_use_voiced_embeddings = kwargs['ap_use_voiced_embeddings']

        if self.decoder_use_unvoiced_bias or self.ap_use_unvoiced_bias:
            assert unvoiced_bias_activation in {'relu', 'exp'}
            self.use_unvoiced_bias = True
            if unvoiced_bias_activation == 'relu':
                unvbias_nonlin = nn.ReLU()
            elif unvoiced_bias_activation == 'exp':
                unvbias_nonlin = ExponentialClass()
            else:
                exit(1)  # we won't reach here anyway due to the assertion
            self.unvoiced_bias_module = nn.Sequential(LinearNorm(n_text_dim, 1), unvbias_nonlin)

        # all situations in which the vpred module is necessary
        if self.ap_use_voiced_embeddings or self.use_unvoiced_bias or 'vpred' in include_modules:
            self.use_vpred_module = True

        if self.use_vpred_module:
            v_model_config['hparams']['n_speaker_dim'] = n_speaker_dim
            self.v_pred_module = get_attribute_prediction_model(v_model_config)
            # 4 embeddings, first two are scales, second two are biases
            if self.ap_use_voiced_embeddings:
                self.v_embeddings = torch.nn.Embedding(4, n_text_dim)
            self.v_pred_threshold = 0.5

        if 'apm' in include_modules:
            f0_model_config['hparams']['n_speaker_dim'] = n_speaker_dim
            energy_model_config['hparams']['n_speaker_dim'] = n_speaker_dim
            if self.use_first_order_features:
                f0_model_config['hparams']['n_in_dim'] = 2
                energy_model_config['hparams']['n_in_dim'] = 2
                if (
                    'spline_flow_params' in f0_model_config['hparams']
                    and f0_model_config['hparams']['spline_flow_params'] is not None
                ):
                    f0_model_config['hparams']['spline_flow_params']['n_in_channels'] = 2
                if (
                    'spline_flow_params' in energy_model_config['hparams']
                    and energy_model_config['hparams']['spline_flow_params'] is not None
                ):
                    energy_model_config['hparams']['spline_flow_params']['n_in_channels'] = 2
            else:
                if (
                    'spline_flow_params' in f0_model_config['hparams']
                    and f0_model_config['hparams']['spline_flow_params'] is not None
                ):
                    f0_model_config['hparams']['spline_flow_params']['n_in_channels'] = f0_model_config['hparams'][
                        'n_in_dim'
                    ]
                if (
                    'spline_flow_params' in energy_model_config['hparams']
                    and energy_model_config['hparams']['spline_flow_params'] is not None
                ):
                    energy_model_config['hparams']['spline_flow_params']['n_in_channels'] = energy_model_config[
                        'hparams'
                    ]['n_in_dim']

            self.f0_pred_module = get_attribute_prediction_model(f0_model_config)
            self.energy_pred_module = get_attribute_prediction_model(energy_model_config)

    def encode_speaker(self, spk_ids):
        spk_ids = spk_ids * 0 if self.dummy_speaker_embedding else spk_ids
        spk_vecs = self.speaker_embedding(spk_ids)
        return spk_vecs

    def encode_text(self, text, in_lens):
        # text_embeddings: b x len_text x n_text_dim
        text_embeddings = self.embedding(text).transpose(1, 2)
        # text_enc: b x n_text_dim x encoder_dim (512)
        text_enc = self.encoder(text_embeddings, in_lens).transpose(1, 2)
        return text_enc, text_embeddings

    def preprocess_context(self, context, speaker_vecs, out_lens, f0, energy_avg, assume_padded=False):
        if self.n_group_size > 1:
            context = self.unfold(context, assume_padded=assume_padded)

            if f0 is not None:
                f0 = self.unfold(f0[:, None, :], assume_padded=assume_padded)
            if energy_avg is not None:
                energy_avg = self.unfold(energy_avg[:, None, :], assume_padded=assume_padded)
        speaker_vecs = speaker_vecs[..., None].expand(-1, -1, context.shape[2])
        context_w_spkvec = torch.cat((context, speaker_vecs), 1)

        if self.use_context_lstm:
            if self.context_lstm_w_f0_and_energy:
                if f0 is not None:
                    context_w_spkvec = torch.cat((context_w_spkvec, f0), 1)

                if energy_avg is not None:
                    context_w_spkvec = torch.cat((context_w_spkvec, energy_avg), 1)

            unfolded_out_lens = out_lens // self.n_group_size
            context_lstm_padded_output = self.context_lstm(context_w_spkvec.transpose(1, 2), unfolded_out_lens)
            context_w_spkvec = context_lstm_padded_output.transpose(1, 2)

        if not self.context_lstm_w_f0_and_energy:
            if f0 is not None:
                context_w_spkvec = torch.cat((context_w_spkvec, f0), 1)

            if energy_avg is not None:
                context_w_spkvec = torch.cat((context_w_spkvec, energy_avg), 1)

        return context_w_spkvec

    def fold(self, mel):
        """Inverse of the self.unfold() operation used for the
        grouping or "squeeze" operation on input

        Args:
            mel: B x C x T tensor of temporal data
        """
        b, d, t = mel.shape
        mel = mel.reshape(b, -1, self.n_group_size, t).transpose(2, 3)
        return mel.reshape(b, -1, t * self.n_group_size)

    def unfold(self, mel, assume_padded=False):
        """operation used for the
        grouping or "squeeze" operation on input

        Args:
            mel: B x C x T tensor of temporal data
        """
        # for inference, mel is being padded beforehand
        if assume_padded:
            b, d, t = mel.shape
            mel = mel.reshape(b, d, -1, self.n_group_size).transpose(2, 3)
            return mel.reshape(b, d * self.n_group_size, -1)
        else:
            return self.unfold_mod(mel.unsqueeze(-1))

    def binarize_attention(self, attn, in_lens, out_lens):
        """For training purposes only. Binarizes attention with MAS. These will
        no longer receive a gradient
        Args:
            attn: B x 1 x max_mel_len x max_text_len
        """
        b_size = attn.shape[0]
        with torch.no_grad():
            attn_cpu = attn.data.cpu().numpy()
            attn_out = torch.zeros_like(attn)
            for ind in range(b_size):
                hard_attn = mas_width1(attn_cpu[ind, 0, : out_lens[ind], : in_lens[ind]])
                attn_out[ind, 0, : out_lens[ind], : in_lens[ind]] = torch.tensor(hard_attn, device=attn.get_device())
        return attn_out

    def get_first_order_features(self, feats, dilation=1):
        """
        feats: b x max_length
        out_lens: b-dim
        """
        # add an extra column
        feats_extended_R = torch.cat((feats, torch.zeros_like(feats[:, 0:dilation])), dim=1)
        feats_extended_L = torch.cat((torch.zeros_like(feats[:, 0:dilation]), feats), dim=1)
        dfeats_R = feats_extended_R[:, dilation:] - feats
        dfeats_L = feats - feats_extended_L[:, 0:-dilation]

        return (dfeats_R + dfeats_L) * 0.5

    def apply_voice_mask_to_text(self, text_enc, voiced_mask):
        """
        text_enc: b x C x N
        voiced_mask: b x N
        """
        voiced_mask = voiced_mask.unsqueeze(1)
        voiced_embedding_s = self.v_embeddings.weight[0:1, :, None]
        unvoiced_embedding_s = self.v_embeddings.weight[1:2, :, None]
        voiced_embedding_b = self.v_embeddings.weight[2:3, :, None]
        unvoiced_embedding_b = self.v_embeddings.weight[3:4, :, None]
        scale = torch.sigmoid(voiced_embedding_s * voiced_mask + unvoiced_embedding_s * (1 - voiced_mask))
        bias = 0.1 * torch.tanh(voiced_embedding_b * voiced_mask + unvoiced_embedding_b * (1 - voiced_mask))
        return text_enc * scale + bias

    def forward(
        self,
        mel,
        speaker_ids,
        text,
        in_lens,
        out_lens,
        binarize_attention=False,
        attn_prior=None,
        f0=None,
        energy_avg=None,
        voiced_mask=None,
    ):
        speaker_vecs = self.encode_speaker(speaker_ids)
        text_enc, text_embeddings = self.encode_text(text, in_lens)

        log_s_list, log_det_W_list, z_mel = [], [], []
        attn_hard = None
        if 'atn' in self.include_modules or 'dec' in self.include_modules:
            # make sure to do the alignments before folding
            attn_mask = ~get_mask_from_lengths(in_lens)[..., None]
            # attn_mask should be 1 for unsd t-steps in text_enc_w_spkvec tensor
            attn_soft, attn_logprob = self.attention(
                mel, text_embeddings, out_lens, attn_mask, key_lens=in_lens, attn_prior=attn_prior
            )

            if binarize_attention:
                attn = self.binarize_attention(attn_soft, in_lens, out_lens)
                attn_hard = attn
            else:
                attn = attn_soft

            context = torch.bmm(text_enc, attn.squeeze(1).transpose(1, 2))
        else:
            raise ValueError(
                f"Something unexpected happened. Both 'atn' and 'dec' are not included in 'self.include_modules'. Please double-check."
            )

        f0_bias = 0
        # unvoiced bias forward pass
        voiced_mask_bool = voiced_mask.bool()
        if self.use_unvoiced_bias:
            f0_bias = self.unvoiced_bias_module(context.permute(0, 2, 1))
            f0_bias = -f0_bias[..., 0]
            f0_bias.masked_fill_(voiced_mask_bool, 0.0)

        # mel decoder forward pass
        if 'dec' in self.include_modules:
            if self.n_group_size > 1:
                # might truncate some frames at the end, but that's ok
                # sometimes referred to as the "squeeze" operation
                # invert this by calling self.fold(mel_or_z)
                mel = self.unfold(mel)
            # where context is folded
            # mask f0 in case values are interpolated
            context_w_spkvec = self.preprocess_context(
                context, speaker_vecs, out_lens, f0 * voiced_mask + f0_bias, energy_avg
            )

            log_s_list, log_det_W_list, z_out = [], [], []
            unfolded_seq_lens = out_lens // self.n_group_size
            for i, flow_step in enumerate(self.flows):
                if i in self.exit_steps:
                    z = mel[:, : self.n_early_size]
                    z_out.append(z)
                    mel = mel[:, self.n_early_size :]
                mel, log_det_W, log_s = flow_step(mel, context_w_spkvec, seq_lens=unfolded_seq_lens)
                log_s_list.append(log_s)
                log_det_W_list.append(log_det_W)

            z_out.append(mel)
            z_mel = torch.cat(z_out, 1)

        # duration predictor forward pass
        duration_model_outputs = None
        if 'dpm' in self.include_modules:
            if attn_hard is None:
                attn_hard = self.binarize_attention(attn_soft, in_lens, out_lens)

            # convert hard attention to durations
            attn_hard_reduced = attn_hard.sum(2)[:, 0, :]
            duration_model_outputs = self.dur_pred_layer(
                torch.detach(text_enc), torch.detach(speaker_vecs), torch.detach(attn_hard_reduced.float()), in_lens
            )

        # f0, energy, vpred predictors forward pass
        f0_model_outputs = None
        energy_model_outputs = None
        vpred_model_outputs = None
        if 'apm' in self.include_modules:
            if attn_hard is None:
                attn_hard = self.binarize_attention(attn_soft, in_lens, out_lens)

            # convert hard attention to durations
            if binarize_attention:
                text_enc_time_expanded = context.clone()
            else:
                text_enc_time_expanded = torch.bmm(text_enc, attn_hard.squeeze(1).transpose(1, 2))

            if self.use_vpred_module:
                # unvoiced bias requires  voiced mask prediction
                vpred_model_outputs = self.v_pred_module(
                    torch.detach(text_enc_time_expanded),
                    torch.detach(speaker_vecs),
                    torch.detach(voiced_mask),
                    out_lens,
                )

                # affine transform context using voiced mask
                if self.ap_use_voiced_embeddings:
                    text_enc_time_expanded = self.apply_voice_mask_to_text(text_enc_time_expanded, voiced_mask)
            if self.ap_use_unvoiced_bias:  # whether to use the unvoiced bias in the attribute predictor
                f0_target = torch.detach(f0 * voiced_mask + f0_bias)
            else:
                f0_target = torch.detach(f0)
            # fit to log f0 in f0 predictor
            f0_target[voiced_mask_bool] = torch.log(f0_target[voiced_mask_bool])
            f0_target = f0_target / 6  # scale to ~ [0, 1] in log space
            energy_avg = energy_avg * 2 - 1  # scale to ~ [-1, 1]

            if self.use_first_order_features:
                df0 = self.get_first_order_features(f0_target)
                denergy_avg = self.get_first_order_features(energy_avg)

                f0_voiced = torch.cat((f0_target[:, None], df0[:, None]), dim=1)
                energy_avg = torch.cat((energy_avg[:, None], denergy_avg[:, None]), dim=1)

                f0_voiced = f0_voiced * 3  # scale to ~ 1 std
                energy_avg = energy_avg * 3  # scale to ~ 1 std
            else:
                f0_voiced = f0_target * 2  # scale to ~ 1 std
                energy_avg = energy_avg * 1.4  # scale to ~ 1 std
            f0_model_outputs = self.f0_pred_module(
                text_enc_time_expanded, torch.detach(speaker_vecs), f0_voiced, out_lens
            )

            energy_model_outputs = self.energy_pred_module(
                text_enc_time_expanded, torch.detach(speaker_vecs), energy_avg, out_lens
            )

        outputs = {
            'z_mel': z_mel,
            'log_det_W_list': log_det_W_list,
            'log_s_list': log_s_list,
            'duration_model_outputs': duration_model_outputs,
            'f0_model_outputs': f0_model_outputs,
            'energy_model_outputs': energy_model_outputs,
            'vpred_model_outputs': vpred_model_outputs,
            'attn_soft': attn_soft,
            'attn': attn,
            'text_embeddings': text_embeddings,
            'attn_logprob': attn_logprob,
        }

        return outputs

    def infer(
        self,
        speaker_id,
        text,
        sigma=0.7,
        speaker_id_text=None,
        speaker_id_attributes=None,
        pace=None,
        token_duration_max=100,
        in_lens=None,
        dur=None,
        f0=None,
        f0_mean=0.0,
        f0_std=0.0,
        energy_avg=None,
        voiced_mask=None,
        pitch_shift=None,
    ):

        batch_size = text.shape[0]
        if in_lens is None:
            in_lens = text.new_ones((batch_size,), dtype=torch.int64) * text.shape[1]
            txt_len_pad_removed = text.shape[1]
        else:
            txt_len_pad_removed = torch.max(in_lens)
            # borisf : this should not be needed as long as we have properly formed input batch
            text = text[:, :txt_len_pad_removed]

        spk_vec = self.encode_speaker(speaker_id)

        if speaker_id_text is None:
            speaker_id_text = speaker_id
        if speaker_id_attributes is None:
            speaker_id_attributes = speaker_id
        spk_vec_text = self.encode_speaker(speaker_id_text)
        spk_vec_attributes = self.encode_speaker(speaker_id_attributes)
        txt_enc, _ = self.encode_text(text, in_lens)

        if dur is None:
            # get token durations
            dur = self.dur_pred_layer.infer(txt_enc, spk_vec_text, lens=in_lens)
            dur = pad_dur(dur, txt_enc)
            dur = dur[:, 0]
            dur = dur.clamp(0, token_duration_max)

        if pace is None:
            pace = txt_enc.new_ones((batch_size, txt_len_pad_removed))
        else:
            pace = pace[:, :txt_len_pad_removed]

        txt_enc_time_expanded, out_lens = regulate_len(
            dur, txt_enc.transpose(1, 2), pace, group_size=self.n_group_size, dur_lens=in_lens,
        )
        n_groups = torch.div(out_lens, self.n_group_size, rounding_mode='floor')
        max_out_len = torch.max(out_lens)

        txt_enc_time_expanded.transpose_(1, 2)
        if voiced_mask is None:
            if self.use_vpred_module:
                # get logits
                voiced_mask = self.v_pred_module.infer(txt_enc_time_expanded, spk_vec_attributes, lens=out_lens)
                voiced_mask_bool = torch.sigmoid(voiced_mask[:, 0]) > self.v_pred_threshold
                voiced_mask = voiced_mask_bool.to(dur.dtype)
            else:
                voiced_mask_bool = None
        else:
            voiced_mask_bool = voiced_mask.bool()

        ap_txt_enc_time_expanded = txt_enc_time_expanded
        # voice mask augmentation only used for attribute prediction
        if self.ap_use_voiced_embeddings:
            ap_txt_enc_time_expanded = self.apply_voice_mask_to_text(txt_enc_time_expanded, voiced_mask)

        f0_bias = 0
        # unvoiced bias forward pass
        if self.use_unvoiced_bias:
            f0_bias = self.unvoiced_bias_module(txt_enc_time_expanded.permute(0, 2, 1))
            f0_bias = -f0_bias[..., 0]

        if f0 is None:
            f0 = self.infer_f0(ap_txt_enc_time_expanded, spk_vec_attributes, voiced_mask_bool, out_lens)[:, 0]

        f0 = adjust_f0(f0, f0_mean, f0_std, voiced_mask_bool, musical_scaling=False)

        if energy_avg is None:
            energy_avg = self.infer_energy(ap_txt_enc_time_expanded, spk_vec, out_lens)[:, 0]

        # replication pad, because ungrouping with different group sizes
        # may lead to mismatched lengths
        # FIXME: use replication pad
        (energy_avg, f0) = pad_energy_avg_and_f0(energy_avg, f0, max_out_len)

        if pitch_shift is not None:
            pitch_shift_spec_len, _ = regulate_len(
                dur,
                pitch_shift[:, :txt_len_pad_removed].unsqueeze(-1),
                pace,
                group_size=self.n_group_size,
                dur_lens=in_lens,
            )
            f0_bias = pitch_shift_spec_len.squeeze(-1) + f0_bias

        context_w_spkvec = self.preprocess_context(
            txt_enc_time_expanded, spk_vec, out_lens, (f0 + f0_bias) * voiced_mask, energy_avg, assume_padded=True,
        )

        residual = txt_enc.new_zeros(batch_size, 80 * self.n_group_size, torch.max(n_groups))
        if sigma > 0.0:
            residual = torch.normal(residual) * sigma

        # map from z sample to data
        num_steps_to_exit = len(self.exit_steps)
        split = num_steps_to_exit * self.n_early_size
        mel = residual[:, split:]
        residual = residual[:, :split]

        for i, flow_step in enumerate(reversed(self.flows)):
            curr_step = self.n_flows - i - 1
            mel = flow_step(mel, context_w_spkvec, inverse=True, seq_lens=n_groups)
            if num_steps_to_exit > 0 and curr_step == self.exit_steps[num_steps_to_exit - 1]:
                # concatenate the next chunk of z
                num_steps_to_exit = num_steps_to_exit - 1
                split = num_steps_to_exit * self.n_early_size
                residual_to_add = residual[:, split:]
                residual = residual[:, :split]
                mel = torch.cat((residual_to_add, mel), 1)

        if self.n_group_size > 1:
            mel = self.fold(mel)

        return {'mel': mel, 'out_lens': out_lens, 'dur': dur, 'f0': f0, 'energy_avg': energy_avg}

    def infer_f0(self, txt_enc_time_expanded, spk_vec, voiced_mask=None, lens=None):
        f0 = self.f0_pred_module.infer(txt_enc_time_expanded, spk_vec, lens)

        # constants
        if self.ap_pred_log_f0:
            if self.use_first_order_features:
                f0 = f0[:, 0:1, :] / 3
            else:
                f0 = f0 / 2
            f0 = f0 * 6
        else:
            f0 = f0 / 6
            f0 = f0 / 640

        if voiced_mask is None:
            voiced_mask = f0 > 0.0
        else:
            if len(voiced_mask.shape) == 2:
                voiced_mask = voiced_mask[:, None]
                # due to grouping, f0 might be 1 frame short
                voiced_mask = voiced_mask[:, :, : f0.shape[-1]]

        if self.ap_pred_log_f0:
            # if variable is set, decoder sees linear f0
            f0 = torch.exp(f0).to(dtype=f0.dtype)
        f0.masked_fill_(~voiced_mask, 0.0)
        return f0

    def infer_energy(self, txt_enc_time_expanded, spk_vec, lens):
        energy = self.energy_pred_module.infer(txt_enc_time_expanded, spk_vec, lens)

        # magic constants
        if self.use_first_order_features:
            energy = energy / 3
        else:
            energy = energy / 1.4
        energy = (energy + 1) / 2
        return energy

    def remove_norms(self):
        """Removes spectral and weightnorms from model. Call before inference
        """
        dev = next(self.parameters()).device
        for name, module in self.named_modules():
            try:
                nn.utils.remove_spectral_norm(module, name='weight_hh_l0')
                print("Removed spectral norm from {}".format(name))
            except:
                pass
            try:
                nn.utils.remove_spectral_norm(module, name='weight_hh_l0_reverse')
                print("Removed spectral norm from {}".format(name))
            except:
                pass
            try:
                nn.utils.remove_weight_norm(module)
                print("Removed wnorm from {}".format(name))
            except:
                pass
        self.to(device=dev)

    @property
    def input_types(self):
        return {
            "text": NeuralType(('B', 'T_text'), TokenIndex()),
            "lens": NeuralType(('B'), LengthsType(), optional=True),
            "speaker_id": NeuralType(('B'), Index()),
            "speaker_id_text": NeuralType(('B'), Index()),
            "speaker_id_attributes": NeuralType(('B'), Index()),
        }

    @property
    def output_types(self):
        return {
            "spect": NeuralType(('B', 'D', 'T_spec'), MelSpectrogramType()),
            "num_frames": NeuralType(('B'), TokenDurationType()),
            "durs_predicted": NeuralType(('B', 'T_text'), TokenDurationType()),
        }



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from unittest import mock

import pytest

from nemo.utils.msc_utils import is_multistorageclient_url


def test_is_multistorageclient_url_with_msc_not_installed():
    with mock.patch('nemo.utils.msc_utils.HAVE_MSC', False):
        assert not is_multistorageclient_url('/tmp/path/to/data.bin')
        assert not is_multistorageclient_url(Path('/tmp/path/to/data.bin'))

        with pytest.raises(ValueError):
            is_multistorageclient_url('msc://profile/path/to/data.bin')


def test_is_multistorageclient_url_with_msc_installed():
    with mock.patch('nemo.utils.msc_utils.HAVE_MSC', True):
        assert is_multistorageclient_url('msc://profile/path/to/data.bin')
        assert not is_multistorageclient_url('/tmp/path/to/data.bin')



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

from nemo.collections.asr.parts.context_biasing.biasing_multi_model import (
    GPUBiasingMultiModel,
    GPUBiasingMultiModelReference,
)
from nemo.collections.asr.parts.context_biasing.boosting_graph_batched import (
    BoostingTreeModelConfig,
    GPUBoostingTreeModel,
)
from nemo.core.utils.optional_libs import TRITON_AVAILABLE

DEVICES = [torch.device("cpu")]
if torch.cuda.is_available():
    DEVICES.append(torch.device("cuda"))
if hasattr(torch, "mps") and torch.mps.is_available():
    DEVICES.append(torch.device("mps"))

# Triton only works on CUDA, so only test use_triton=True if Triton is available
USE_TRITON_OPTIONS = [False, True] if TRITON_AVAILABLE else [False]


def create_boosting_model(phrases: list[str], tokenizer, device: torch.device) -> GPUBoostingTreeModel:
    """Helper to create boosting model from phrases"""
    cfg = BoostingTreeModelConfig(key_phrases_list=phrases, context_score=1.0)
    model = GPUBoostingTreeModel.from_config(cfg, tokenizer=tokenizer)
    return model.to(device)


class TestGPUBiasingMultiModel:
    @pytest.mark.unit
    @pytest.mark.with_downloads
    @pytest.mark.parametrize("device", DEVICES)
    def test_add_models_incremental(self, stt_en_conformer_transducer_small, device: torch.device):
        """Test adding 2 boosting models one-by-one, verifying arcs and states are correctly merged."""
        tokenizer = stt_en_conformer_transducer_small.tokenizer
        vocab_size = tokenizer.vocab_size

        # Create empty multi-model
        multi_model = GPUBiasingMultiModel(vocab_size=vocab_size).to(device)

        # Initially empty
        assert multi_model.num_models == 0
        assert multi_model.has_models() is False
        assert multi_model.num_states_total == 0
        assert multi_model.num_arcs_extended_total == 0

        # Create and add first model
        model1 = create_boosting_model(["hello", "world"], tokenizer, device)
        model_id1 = multi_model.add_model(model1, alpha=1.0)

        # Verify after first model
        assert model_id1 == 0
        assert multi_model.num_models == 1
        assert multi_model.has_models() is True
        assert multi_model.model2active[model_id1].item() is True
        assert multi_model.num_states_total == model1.num_states
        assert multi_model.num_arcs_extended_total == model1.num_arcs_extended
        assert multi_model.model2num_states[model_id1].item() == model1.num_states
        assert multi_model.model2num_arcs_extended[model_id1].item() == model1.num_arcs_extended

        # Create and add second model
        model2 = create_boosting_model(["test", "one", "two"], tokenizer, device)
        model_id2 = multi_model.add_model(model2, alpha=1.5)

        # Verify after second model
        assert model_id2 == 1
        assert multi_model.num_models == 2
        assert multi_model.has_models() is True
        assert multi_model.model2active[model_id1].item() is True
        assert multi_model.model2active[model_id2].item() is True
        assert multi_model.num_states_total == model1.num_states + model2.num_states
        assert multi_model.num_arcs_extended_total == model1.num_arcs_extended + model2.num_arcs_extended

        # Verify offsets
        assert multi_model.model2states_offset[model_id1].item() == 0
        assert multi_model.model2states_offset[model_id2].item() == model1.num_states
        assert multi_model.model2arcs_offset[model_id1].item() == 0
        assert multi_model.model2arcs_offset[model_id2].item() == model1.num_arcs_extended

        # Verify init states work
        init_states = multi_model.get_init_states(batch_size=4, bos=True)
        assert init_states.shape == (4,)
        assert init_states.device.type == device.type

    @pytest.mark.unit
    @pytest.mark.with_downloads
    @pytest.mark.parametrize("device", DEVICES)
    def test_add_then_remove_model(self, stt_en_conformer_transducer_small, device: torch.device):
        """Test adding 2 models then removing the first one."""
        tokenizer = stt_en_conformer_transducer_small.tokenizer
        vocab_size = tokenizer.vocab_size

        multi_model = GPUBiasingMultiModel(vocab_size=vocab_size).to(device)

        # Add two models
        model1 = create_boosting_model(["alpha", "beta"], tokenizer, device)
        model2 = create_boosting_model(["gamma", "delta"], tokenizer, device)

        model_id1 = multi_model.add_model(model1, alpha=1.0)
        model_id2 = multi_model.add_model(model2, alpha=2.0)

        # Store counts before removal
        model1_num_states = model1.num_states
        model1_num_arcs = model1.num_arcs_extended
        total_states_before = multi_model.num_states_total
        total_arcs_before = multi_model.num_arcs_extended_total

        assert multi_model.model2active[model_id1].item() is True
        assert multi_model.model2active[model_id2].item() is True

        # Remove first model
        multi_model.remove_model(model_id1)

        # Verify removal
        assert model_id1 in multi_model.free_ids
        assert multi_model.model2active[model_id1].item() is False
        assert multi_model.model2active[model_id2].item() is True
        assert multi_model.model2alpha[model_id1].item() == 0.0
        assert multi_model.model2alpha[model_id2].item() == 2.0

        # Verify state/arc counts decreased
        assert multi_model.num_states_total == total_states_before - model1_num_states
        assert multi_model.num_arcs_extended_total == total_arcs_before - model1_num_arcs

        # Verify model2 offset updated (shifted left)
        assert multi_model.model2states_offset[model_id2].item() == 0
        assert multi_model.model2arcs_offset[model_id2].item() == 0

    @pytest.mark.unit
    @pytest.mark.with_downloads
    @pytest.mark.parametrize("device", DEVICES)
    def test_model_id_reuse(self, stt_en_conformer_transducer_small, device):
        """Test that removed model IDs are reused."""
        tokenizer = stt_en_conformer_transducer_small.tokenizer
        vocab_size = tokenizer.vocab_size

        multi_model = GPUBiasingMultiModel(vocab_size=vocab_size).to(device)

        # Add model1 -> id=0
        model1 = create_boosting_model(["first"], tokenizer, device)
        model_id1 = multi_model.add_model(model1)
        assert model_id1 == 0

        # Add model2 -> id=1
        model2 = create_boosting_model(["second"], tokenizer, device)
        model_id2 = multi_model.add_model(model2)
        assert model_id2 == 1

        # Remove model1
        multi_model.remove_model(model_id1)
        assert model_id1 in multi_model.free_ids

        # Add model3 -> should reuse id=0
        model3 = create_boosting_model(["third"], tokenizer, device)
        model_id3 = multi_model.add_model(model3)
        assert model_id3 == model_id1  # Reused ID
        assert model_id1 not in multi_model.free_ids  # No longer free

        # Verify model3 is active
        assert multi_model.model2active[model_id3].item() is True

    @pytest.mark.unit
    @pytest.mark.with_downloads
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("batch_size", [1, 4])
    @pytest.mark.parametrize("use_triton", USE_TRITON_OPTIONS)
    @pytest.mark.parametrize("bos", [True, False])
    def test_advance_matches_reference(
        self, stt_en_conformer_transducer_small, device: torch.device, batch_size: int, use_triton, bos: bool
    ):
        """Verify GPUBiasingMultiModel produces same scores/states as reference implementation."""
        tokenizer = stt_en_conformer_transducer_small.tokenizer
        vocab_size = tokenizer.vocab_size

        # Create both implementations
        multi_model = GPUBiasingMultiModel(vocab_size=vocab_size, use_triton=use_triton).to(device)
        reference = GPUBiasingMultiModelReference(vocab_size=vocab_size).to(device)

        # Create boosting models with same phrases
        phrases1 = ["hello world", "test"]
        phrases2 = ["neural", "network"]

        model1_mm = create_boosting_model(phrases1, tokenizer, device)
        model1_ref = create_boosting_model(phrases1, tokenizer, device)
        model2_mm = create_boosting_model(phrases2, tokenizer, device)
        model2_ref = create_boosting_model(phrases2, tokenizer, device)

        # Add models to both with same alpha values
        alpha1, alpha2 = 1.0, 1.5
        model_id1_mm = multi_model.add_model(model1_mm, alpha=alpha1)
        model_id1_ref = reference.add_model(model1_ref, alpha=alpha1)
        model_id2_mm = multi_model.add_model(model2_mm, alpha=alpha2)
        model_id2_ref = reference.add_model(model2_ref, alpha=alpha2)

        assert model_id1_mm == model_id1_ref
        assert model_id2_mm == model_id2_ref

        # Get initial states
        states_mm = multi_model.get_init_states(batch_size=batch_size, bos=bos)
        states_ref = reference.get_init_states(batch_size=batch_size, bos=bos)

        # Create model_ids tensor with alternating models
        model_ids = torch.tensor(
            [model_id1_mm if i % 2 == 0 else model_id2_mm for i in range(batch_size)],
            dtype=torch.long,
            device=device,
        )

        # Call advance on both
        scores_mm, next_states_mm = multi_model.advance(states_mm, model_ids)
        scores_ref, next_states_ref = reference.advance(states_ref, model_ids)

        # Verify shapes
        assert scores_mm.shape == (batch_size, vocab_size)
        assert next_states_mm.shape == (batch_size, vocab_size)
        assert scores_ref.shape == (batch_size, vocab_size)
        assert next_states_ref.shape == (batch_size, vocab_size)

        # Verify scores and states match
        assert torch.allclose(
            scores_mm, scores_ref, atol=1e-5
        ), f"Scores mismatch: max diff = {(scores_mm - scores_ref).abs().max()}"
        assert torch.equal(next_states_mm, next_states_ref), "Next states mismatch"

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_empty_multi_model(self, device: torch.device):
        """Test behavior of empty multi-model."""
        vocab_size = 100
        multi_model = GPUBiasingMultiModel(vocab_size=vocab_size, use_triton=False).to(device)

        # Verify empty state
        assert multi_model.has_models() is False
        assert multi_model.num_models == 0
        assert multi_model.num_states_total == 0
        assert multi_model.num_arcs_extended_total == 0

        # get_init_states should work and return START_STATE
        init_states = multi_model.get_init_states(batch_size=4, bos=True)
        assert init_states.shape == (4,)
        assert (init_states == GPUBiasingMultiModel.START_STATE).all()



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pytest
import torch
from examples.asr.asr_cache_aware_streaming.speech_to_text_multitalker_streaming_infer import (
    MultitalkerTranscriptionConfig,
)
from omegaconf import OmegaConf

from nemo.collections.asr.models.configs.asr_models_config import CacheAwareStreamingConfig
from nemo.collections.asr.parts.utils.multispk_transcribe_utils import (
    MultiTalkerInstanceManager,
    SpeakerTaggedASR,
    append_word_and_ts_seq,
    fix_frame_time_step,
    get_multi_talker_samples_from_manifest,
    get_multitoken_words,
    get_new_sentence_dict,
    get_simulated_softmax,
    get_word_dict_content_offline,
    get_word_dict_content_online,
    write_seglst,
)
from tests.collections.asr.test_asr_rnnt_encoder_model_bpe import asr_model as offline_asr_model
from tests.collections.speaker_tasks.test_diar_sortformer_models import sortformer_model as diar_model


@pytest.fixture()
def asr_model(offline_asr_model):
    """Wrapper fixture that adds streaming_cfg to the asr_model from test_asr_rnnt_encoder_model_bpe"""
    # Add streaming_cfg to encoder for streaming tests
    streaming_cfg = CacheAwareStreamingConfig(
        valid_out_len=1,
        drop_extra_pre_encoded=7,
        chunk_size=8,
        shift_size=4,
        cache_drop_size=4,
        last_channel_cache_size=64,
        pre_encode_cache_size=0,
        last_channel_num=0,
        last_time_num=0,
    )
    offline_asr_model.encoder.streaming_cfg = streaming_cfg

    # Mock get_initial_cache_state method for MultiTalkerInstanceManager tests
    def get_initial_cache_state(batch_size=1):
        """Mock method to return initial cache state for streaming"""
        # Return dummy cache state tensors
        cache_last_channel = torch.zeros(2, batch_size, 64)
        cache_last_time = torch.zeros(2, batch_size, 64)
        cache_last_channel_len = torch.zeros(batch_size)
        return (cache_last_channel, cache_last_time, cache_last_channel_len)

    offline_asr_model.encoder.get_initial_cache_state = get_initial_cache_state

    return offline_asr_model


class TestGetNewSentenceDict:
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "speaker,start_time,end_time,text,session_id",
        [
            ("speaker_0", 0.0, 1.0, "hello", None),
            ("spk1", 1.23, 4.56, "world", "session_A"),
        ],
    )
    def test_get_new_sentence_dict(self, speaker, start_time, end_time, text, session_id):
        result = get_new_sentence_dict(
            speaker=speaker, start_time=start_time, end_time=end_time, text=text, session_id=session_id
        )
        assert result["speaker"] == speaker
        assert result["start_time"] == start_time
        assert result["end_time"] == end_time
        assert result["words"] == text.lstrip()
        assert result["session_id"] == session_id


class TestFixFrameTimeStep:
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "new_tokens,new_words,frame_inds_seq,expected",
        [
            # Case 1: Trim when frame_inds longer than tokens
            (["x", "y"], ["x", "y"], [1, 2, 3, 4], [3, 4]),
            # Case 2: No change when lengths match
            (["u", "v", "w"], ["u", "v", "w"], [7, 8, 9], [7, 8, 9]),
        ],
    )
    def test_fix_frame_time_step_shapes(self, new_tokens, new_words, frame_inds_seq, expected):
        cfg = OmegaConf.structured(MultitalkerTranscriptionConfig(log=False))
        out = fix_frame_time_step(cfg, new_tokens, new_words, frame_inds_seq)
        assert out == expected


class TestGetSimulatedSoftmax:
    @pytest.mark.unit
    def test_invalid_dims(self):
        cfg = OmegaConf.structured(MultitalkerTranscriptionConfig(min_sigmoid_val=0.0, max_num_of_spks=2))
        with pytest.raises(ValueError):
            get_simulated_softmax(cfg, torch.zeros((2, 3)))

    @pytest.mark.unit
    def test_invalid_length_vs_maxspks(self):
        cfg = OmegaConf.structured(MultitalkerTranscriptionConfig(min_sigmoid_val=0.0, max_num_of_spks=4))
        with pytest.raises(ValueError):
            get_simulated_softmax(cfg, torch.zeros(3))

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "vec,min_sigmoid_val,max_num_of_spks,expected_prefix",
        [
            # Clamp first element to min, normalize across all then zero out tail
            ([0.0, 0.2, 0.3, 0.5], 0.1, 3, [0.1 / 1.1, 0.2 / 1.1, 0.3 / 1.1]),
            # Zero-sum uniform case with tail zeroed
            ([0.0, 0.0, 0.0, 0.0], 0.0, 2, [0.25, 0.25]),
        ],
    )
    def test_valid_softmax_behavior(self, vec, min_sigmoid_val, max_num_of_spks, expected_prefix):
        cfg = OmegaConf.structured(
            MultitalkerTranscriptionConfig(min_sigmoid_val=min_sigmoid_val, max_num_of_spks=max_num_of_spks)
        )
        out = get_simulated_softmax(cfg, torch.tensor(vec))
        assert out.shape[0] == len(vec)
        # Tail past max_num_of_spks is zero
        if len(vec) > max_num_of_spks:
            assert torch.all(out[max_num_of_spks:] == 0)
        # First max_num_of_spks entries match expected prefix approximately
        assert torch.allclose(out[: len(expected_prefix)], torch.tensor(expected_prefix), atol=1e-5)


class TestWordDictContentOffline:
    @pytest.mark.unit
    @pytest.mark.parametrize("frame_stt,frame_end,expected_end", [(2, 5, 5), (2, 2, 3)])
    def test_get_word_dict_content_offline(self, frame_stt, frame_end, expected_end):
        # diar_pred_out with highest mean on speaker 2 within the selected frames
        T, N = 6, 3
        diar_pred_out = torch.zeros((T, N))
        diar_pred_out[2:5, 2] = 10.0
        cfg = OmegaConf.structured(
            MultitalkerTranscriptionConfig(
                left_frame_shift=0, right_frame_shift=0, max_num_of_spks=3, min_sigmoid_val=0.0
            )
        )
        word = "hello"
        res = get_word_dict_content_offline(
            cfg=cfg,
            word=word,
            word_index=0,
            diar_pred_out=diar_pred_out,
            time_stt_end_tuple=(frame_stt, frame_end),
            frame_len=0.08,
        )
        assert res["word"] == word
        assert res["speaker"] == "speaker_2"
        assert res["frame_stt"] == frame_stt
        assert res["frame_end"] == expected_end
        assert abs(res["start_time"] - frame_stt * 0.08) < 1e-6
        assert abs(res["end_time"] - expected_end * 0.08) < 1e-6


class TestWordDictContentOnline:
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "token_group,frame_inds_seq,time_step_local_offset,expected_stt,expected_end",
        [
            # Single token: end = stt + 1
            (["t1"], [4, 5, 6], 1, 5, 6),
            # Multi-token falling back due to IndexError -> end = stt + 1
            (["t1", "t2", "t3"], [8], 0, 8, 9),
        ],
    )
    def test_get_word_dict_content_online(
        self, token_group, frame_inds_seq, time_step_local_offset, expected_stt, expected_end
    ):
        T, N = 12, 4
        diar_pred_out_stream = torch.zeros((T, N))
        diar_pred_out_stream[expected_stt:expected_end, 1] = 5.0  # Make speaker 1 dominant in the window
        cfg = OmegaConf.structured(
            MultitalkerTranscriptionConfig(
                left_frame_shift=0, right_frame_shift=0, max_num_of_spks=4, min_sigmoid_val=0.0
            )
        )

        res = get_word_dict_content_online(
            cfg=cfg,
            word="token",
            word_index=0,
            diar_pred_out_stream=diar_pred_out_stream,
            token_group=token_group,
            frame_inds_seq=frame_inds_seq,
            time_step_local_offset=time_step_local_offset,
            frame_len=0.08,
        )
        assert res["frame_stt"] == expected_stt
        assert res["frame_end"] == expected_end
        assert res["speaker"] == "speaker_1"


class TestGetMultitokenWords:
    @pytest.mark.unit
    @pytest.mark.parametrize("verbose", [True, False])
    def test_get_multitoken_words_replaces_shorter_saved(self, verbose):
        cfg = OmegaConf.structured(MultitalkerTranscriptionConfig(verbose=verbose))
        word_and_ts_seq = {
            "words": [{"word": "hello"}, {"word": "multi"}, {"word": "world"}],
            "buffered_words": [],
            "word_window_seq": [],
        }
        word_seq = ["hello", "multi token", "world", "new"]
        new_words = ["new"]
        out = get_multitoken_words(cfg, word_and_ts_seq, word_seq, new_words, fix_prev_words_count=2)
        # The second-to-last element should be replaced by the longer previous word
        assert out["words"][1]["word"] == "multi token"


class TestAppendWordAndTsSeq:
    @pytest.mark.unit
    def test_append_and_fifo_pop(self):
        cfg = OmegaConf.structured(MultitalkerTranscriptionConfig(word_window=2))
        word_and_ts_seq = {
            "words": [{"word": "a", "speaker": "speaker_0"}, {"word": "b", "speaker": "speaker_1"}],
            "buffered_words": [{"word": "a", "speaker": "speaker_0"}, {"word": "b", "speaker": "speaker_1"}],
            "token_frame_index": [],
            "offset_count": 0,
            "status": "success",
            "sentences": None,
            "last_word_index": 0,
            "speaker_count": None,
            "transcription": None,
            "max_spk_probs": [],
            "word_window_seq": ["a", "b"],
            "speaker_count_buffer": ["speaker_0", "speaker_1"],
            "sentence_memory": {},
        }
        word_dict = {"word": "c", "speaker": "speaker_1"}
        word_idx_offset, out = append_word_and_ts_seq(cfg, 0, word_and_ts_seq, word_dict)
        assert word_idx_offset == 0
        # FIFO: buffered_words and word_window_seq should maintain length <= word_window
        assert len(out["buffered_words"]) == cfg.word_window
        assert len(out["word_window_seq"]) == cfg.word_window
        # speaker_count: unique speakers in buffer
        assert out["speaker_count"] == 2


class TestGetDiarPredOutStream:
    class Dummy:
        def __init__(self, diar_model, nframes):
            self.diar_model = diar_model
            self._nframes_per_chunk = nframes

    @pytest.mark.unit
    @pytest.mark.parametrize("step_num,nframes", [(0, 3), (1, 3)])
    def test_get_diar_pred_out_stream(self, diar_model, step_num, nframes):
        B, T, N = 2, 10, 4
        mats = torch.arange(B * T * N, dtype=torch.float32).reshape(B, T, N)
        # Set rttms_mask_mats on the diar_model
        diar_model.rttms_mask_mats = mats
        dummy = self.Dummy(diar_model, nframes)
        new_stream, new_chunk = SpeakerTaggedASR.get_diar_pred_out_stream(dummy, step_num)

        start = step_num * nframes
        end = start + nframes
        assert new_stream.shape[1] == min(end, T)
        assert torch.equal(new_chunk, new_stream[:, start:end])


class TestWriteSeglst:
    @pytest.mark.unit
    def test_write_and_read(self, tmp_path):
        seglst = [
            {"speaker": "speaker_0", "start_time": 0.0, "end_time": 1.0, "words": "hi", "session_id": "S1"},
            {"speaker": "speaker_1", "start_time": 1.0, "end_time": 2.0, "words": "there", "session_id": "S1"},
        ]
        outpath = tmp_path / "out.json"
        write_seglst(str(outpath), seglst)
        content = outpath.read_text(encoding="utf-8")
        assert content == json.dumps(seglst, indent=2) + "\n"


class TestGetMultiTalkerSamplesFromManifest:
    @pytest.mark.unit
    def test_missing_audio_filepath(self, tmp_path):
        mpath = tmp_path / "manifest.jsonl"
        mpath.write_text(json.dumps({}) + "\n", encoding="utf-8")
        cfg = OmegaConf.structured(MultitalkerTranscriptionConfig(spk_supervision="none"))
        with pytest.raises(KeyError):
            get_multi_talker_samples_from_manifest(cfg, str(mpath), feat_per_sec=100.0, max_spks=2)

    @pytest.mark.unit
    def test_rttm_missing_file(self, tmp_path):
        mpath = tmp_path / "manifest.jsonl"
        missing_rttm = str(tmp_path / "missing.rttm")
        line = {
            "audio_filepath": "sample.wav",
            "duration": 10.0,
            "rttm_filepath": missing_rttm,
        }
        mpath.write_text(json.dumps(line) + "\n", encoding="utf-8")
        cfg = OmegaConf.structured(MultitalkerTranscriptionConfig(spk_supervision="rttm"))
        with pytest.raises(FileNotFoundError):
            get_multi_talker_samples_from_manifest(cfg, str(mpath), feat_per_sec=100.0, max_spks=2)


class TestSpeakerTaggedASRInit:
    """Test the initialization of SpeakerTaggedASR class"""

    @pytest.mark.unit
    def test_init_default_config_values(self, asr_model, diar_model, tmp_path):
        """Test initialization with default config values using .get() from MultitalkerTranscriptionConfig"""
        audio_path = tmp_path / "test.wav"
        audio_path.touch()

        cfg = MultitalkerTranscriptionConfig(
            manifest_file=None,
            audio_file=str(audio_path),
            fix_prev_words_count=5,
            update_prev_words_sentence=10,
            ignored_initial_frame_steps=2,
            max_num_of_spks=3,
            att_context_size=[0, 40],
            binary_diar_preds=True,
            batch_size=1,
            sent_break_sec=30.0,  # Explicit value
            masked_asr=True,  # Explicit value
            cache_gating=False,  # Explicit value
            mask_preencode=False,  # Explicit value
            single_speaker_mode=False,  # Explicit value
            generate_realtime_scripts=False,
        )
        # Convert to OmegaConf to support .get() method
        cfg = OmegaConf.structured(cfg)

        speaker_tagged_asr = SpeakerTaggedASR(cfg, asr_model, diar_model)

        # Verify values from .get() calls are properly set
        # pylint: disable=protected-access
        assert speaker_tagged_asr._max_num_of_spks == 3  # From cfg.get("max_num_of_spks", 4)
        assert speaker_tagged_asr._sent_break_sec == 30.0  # From cfg
        # cache_gating and cache_gating_buffer_size use defaults via cfg.get() since they're not in MultitalkerTranscriptionConfig
        assert speaker_tagged_asr._cache_gating is False  # Default value from cfg.get("cache_gating", False)
        assert (
            speaker_tagged_asr._cache_gating_buffer_size == 2
        )  # Default value from cfg.get("cache_gating_buffer_size", 2)
        assert speaker_tagged_asr._masked_asr is True  # From cfg
        assert speaker_tagged_asr._use_mask_preencode is False  # From cfg
        assert speaker_tagged_asr._single_speaker_mode is False  # From cfg
        # pylint: enable=protected-access

    @pytest.mark.unit
    def test_init_instance_manager_creation(self, asr_model, diar_model, tmp_path):
        """Test that instance_manager is properly created during initialization"""
        audio_path = tmp_path / "test.wav"
        audio_path.touch()

        cfg = MultitalkerTranscriptionConfig(
            manifest_file=None,
            audio_file=str(audio_path),
            fix_prev_words_count=5,
            update_prev_words_sentence=10,
            ignored_initial_frame_steps=2,
            max_num_of_spks=4,
            att_context_size=[0, 50],
            binary_diar_preds=True,
            batch_size=2,
            generate_realtime_scripts=False,
        )
        # Convert to OmegaConf to support .get() method
        cfg = OmegaConf.structured(cfg)

        speaker_tagged_asr = SpeakerTaggedASR(cfg, asr_model, diar_model)

        # Verify instance_manager is created and has correct attributes
        assert speaker_tagged_asr.instance_manager is not None
        assert isinstance(speaker_tagged_asr.instance_manager, MultiTalkerInstanceManager)
        assert speaker_tagged_asr.instance_manager.asr_model == asr_model
        assert speaker_tagged_asr.instance_manager.diar_model == diar_model
        assert speaker_tagged_asr.instance_manager.max_num_of_spks == 4
        assert speaker_tagged_asr.instance_manager.batch_size == 2


class TestSpeakerTaggedASRMethods:
    """Test various methods of the SpeakerTaggedASR class"""

    @pytest.mark.unit
    def test_get_offset_sentence(self, asr_model, diar_model, tmp_path):
        """Test _get_offset_sentence method"""
        audio_path = tmp_path / "test.wav"
        audio_path.touch()

        cfg = MultitalkerTranscriptionConfig(
            manifest_file=None,
            audio_file=str(audio_path),
            fix_prev_words_count=5,
            update_prev_words_sentence=10,
            ignored_initial_frame_steps=2,
            max_num_of_spks=4,
            att_context_size=[0, 50],
            binary_diar_preds=True,
            batch_size=1,
            generate_realtime_scripts=False,
        )
        cfg = OmegaConf.structured(cfg)

        speaker_tagged_asr = SpeakerTaggedASR(cfg, asr_model, diar_model)

        # Create a mock session_trans_dict
        session_trans_dict = {
            'uniq_id': 'session_1',
            'words': [
                {'speaker': 'speaker_0', 'start_time': 0.0, 'end_time': 0.5, 'word': 'hello'},
                {'speaker': 'speaker_0', 'start_time': 0.5, 'end_time': 1.0, 'word': 'world'},
            ],
        }

        # pylint: disable=protected-access
        result = speaker_tagged_asr._get_offset_sentence(session_trans_dict, 0)
        # pylint: enable=protected-access

        assert result['session_id'] == 'session_1'
        assert result['speaker'] == 'speaker_0'
        assert result['start_time'] == 0.0
        assert result['end_time'] == 0.5
        assert result['words'] == 'hello '

    @pytest.mark.unit
    def test_find_active_speakers_valid(self, asr_model, diar_model, tmp_path):
        """Test _find_active_speakers with valid inputs"""
        audio_path = tmp_path / "test.wav"
        audio_path.touch()

        cfg = MultitalkerTranscriptionConfig(
            manifest_file=None,
            audio_file=str(audio_path),
            fix_prev_words_count=5,
            update_prev_words_sentence=10,
            ignored_initial_frame_steps=2,
            max_num_of_spks=4,
            att_context_size=[0, 50],
            binary_diar_preds=True,
            batch_size=1,
            generate_realtime_scripts=False,
        )
        cfg = OmegaConf.structured(cfg)

        speaker_tagged_asr = SpeakerTaggedASR(cfg, asr_model, diar_model)

        # Create mock diar predictions: (B=2, T=10, N=4)
        diar_preds = torch.zeros(2, 10, 4)
        # First batch: speakers 0 and 2 are active (high values)
        diar_preds[0, :, 0] = 0.8
        diar_preds[0, :, 2] = 0.9
        # Second batch: speaker 1 is active
        diar_preds[1, :, 1] = 0.7

        # pylint: disable=protected-access
        result = speaker_tagged_asr._find_active_speakers(diar_preds, n_active_speakers_per_stream=2)
        # pylint: enable=protected-access

        assert len(result) == 2
        assert 0 in result[0] and 2 in result[0]
        assert 1 in result[1]

    @pytest.mark.unit
    def test_mask_features_valid(self, asr_model, diar_model, tmp_path):
        """Test mask_features with valid inputs"""
        audio_path = tmp_path / "test.wav"
        audio_path.touch()

        cfg = MultitalkerTranscriptionConfig(
            manifest_file=None,
            audio_file=str(audio_path),
            fix_prev_words_count=5,
            update_prev_words_sentence=10,
            ignored_initial_frame_steps=2,
            max_num_of_spks=4,
            att_context_size=[0, 50],
            binary_diar_preds=True,
            batch_size=1,
            generate_realtime_scripts=False,
        )
        cfg = OmegaConf.structured(cfg)

        speaker_tagged_asr = SpeakerTaggedASR(cfg, asr_model, diar_model)

        # Create mock audio: (B=2, C=80, T=100)
        chunk_audio = torch.randn(2, 80, 100)
        # Create mask: (B=2, T=12) - will be expanded to match T=100
        mask = torch.zeros(2, 12)
        mask[0, :5] = 0.8  # First batch: first 5 frames active
        mask[1, 5:] = 0.9  # Second batch: last 7 frames active

        result = speaker_tagged_asr.mask_features(chunk_audio, mask, threshold=0.5, mask_value=-16.6355)

        assert result.shape == chunk_audio.shape
        assert result.dtype == chunk_audio.dtype



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import json
import tempfile
from pathlib import Path

import numpy as np
import pytest
import torch

from nemo.collections.tts.parts.preprocessing.feature_processors import (
    FeatureScaler,
    LogCompression,
    MeanVarianceNormalization,
    MeanVarianceSpeakerNormalization,
)


class TestTTSFeatureProcessors:
    @contextlib.contextmanager
    def _write_test_dict(self, test_dict, filename):
        temp_dir = tempfile.TemporaryDirectory()
        try:
            test_dir = Path(temp_dir.name)
            test_dict_filepath = test_dir / filename
            with open(test_dict_filepath, 'w', encoding="utf-8") as stats_f:
                json.dump(test_dict, stats_f, indent=4)

            yield test_dict_filepath
        finally:
            temp_dir.cleanup()

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_feature_scalar(self):
        field = "test_feat"
        input_tensor = torch.tensor([-2.5, 0.0, 1.0], dtype=torch.float32)
        expected_tensor = torch.tensor([0.0, 2.0, 2.8], dtype=torch.float32)
        processor = FeatureScaler(field, add_value=2.5, div_value=1.25)

        training_example = {field: input_tensor}
        processor.process(training_example)
        output_tensor = training_example[field]

        torch.testing.assert_close(output_tensor, expected_tensor)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_log_compression(self):
        field = "test_feat"

        input_tensor = torch.tensor([-0.5, 0.0, 2.0], dtype=torch.float32)
        expected_tensor = torch.tensor([np.log(0.5), 0.0, np.log(3.0)], dtype=torch.float32)
        processor = LogCompression(field)

        training_example = {field: input_tensor}
        processor.process(training_example)
        output_tensor = training_example[field]

        torch.testing.assert_close(output_tensor, expected_tensor)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_log_compression_clamp(self):
        field = "test_feat"

        input_tensor = torch.tensor([0.1, 1.0, 2.0], dtype=torch.float32)
        expected_tensor = torch.tensor([np.log(0.5), 0.0, np.log(2.0)], dtype=torch.float32)
        processor = LogCompression(field, log_zero_guard_type="clamp", log_zero_guard_value=0.5)

        training_example = {field: input_tensor}
        processor.process(training_example)
        output_tensor = training_example[field]

        torch.testing.assert_close(output_tensor, expected_tensor)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_mean_variance_normalization(self):
        field = "test_feat"
        filename = "stats.json"
        stat_dict = {"default": {"test_feat_mean": 1.5, "test_feat_std": 0.5}}

        input_tensor = torch.tensor([0.0, 1.5, 2.0], dtype=torch.float32)
        expected_tensor = torch.tensor([-3.0, 0.0, 1.0], dtype=torch.float32)
        training_example = {field: input_tensor}

        with self._write_test_dict(stat_dict, filename=filename) as stat_dict_filepath:
            processor = MeanVarianceNormalization(field, stats_path=stat_dict_filepath, mask_field=None)
            processor.process(training_example)

        output_tensor = training_example[field]
        torch.testing.assert_close(output_tensor, expected_tensor)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_mean_variance_normalization_masked(self):
        field = "test_feat"
        mask_field = "mask"
        filename = "stats.json"
        stat_dict = {"default": {"test_feat_mean": 1.0, "test_feat_std": 0.5}}

        input_tensor = torch.tensor([2.0, 3.0, 4.0, 5.0], dtype=torch.float32)
        input_mask = torch.tensor([True, False, False, True], dtype=torch.bool)
        expected_tensor = torch.tensor([2.0, 0.0, 0.0, 8.0], dtype=torch.float32)
        training_example = {field: input_tensor, mask_field: input_mask}

        with self._write_test_dict(stat_dict, filename=filename) as stat_dict_filepath:
            processor = MeanVarianceNormalization(field, stats_path=stat_dict_filepath, mask_field=mask_field)
            processor.process(training_example)

        output_tensor = training_example[field]
        torch.testing.assert_close(output_tensor, expected_tensor)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_mean_variance_speaker_normalization(self):
        field = "pitch"
        filename = "stats.json"
        stat_dict = {
            "default": {"pitch_mean": 1.5, "pitch_std": 0.5},
            "speaker1": {"pitch_mean": 0.5, "pitch_std": 1.0},
            "speaker2": {"pitch_mean": 0.0, "pitch_std": 2.0},
        }

        input_tensor = torch.tensor([0.0, 1.0], dtype=torch.float32)

        training_example1 = {field: input_tensor, "speaker": "speaker1"}
        training_example2 = {field: input_tensor, "speaker": "speaker2"}
        training_example3 = {field: input_tensor, "speaker": "unknown"}
        expected_tensor1 = torch.tensor([-0.5, 0.5], dtype=torch.float32)
        expected_tensor2 = torch.tensor([0.0, 0.5], dtype=torch.float32)
        expected_tensor3 = torch.tensor([-3.0, -1.0], dtype=torch.float32)

        with self._write_test_dict(stat_dict, filename=filename) as stat_dict_filepath:
            processor = MeanVarianceSpeakerNormalization(
                field, stats_path=stat_dict_filepath, mask_field=None, fallback_to_default=True
            )
            processor.process(training_example1)
            processor.process(training_example2)
            processor.process(training_example3)

        output_tensor1 = training_example1[field]
        output_tensor2 = training_example2[field]
        output_tensor3 = training_example3[field]
        torch.testing.assert_close(output_tensor1, expected_tensor1)
        torch.testing.assert_close(output_tensor2, expected_tensor2)
        torch.testing.assert_close(output_tensor3, expected_tensor3)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_mean_variance_speaker_normalization_masked(self):
        field = "test_feat"
        mask_field = "test_mask"
        filename = "stats.json"
        stat_dict = {"steve": {"test_feat_mean": -1.0, "test_feat_std": 2.0}}

        input_tensor = torch.tensor([1.0, 2.0, 3.0, 4.0], dtype=torch.float32)
        input_mask = torch.tensor([False, True, False, True], dtype=torch.bool)
        expected_tensor = torch.tensor([0.0, 1.5, 0.0, 2.5], dtype=torch.float32)

        training_example = {field: input_tensor, "speaker": "steve", mask_field: input_mask}

        with self._write_test_dict(stat_dict, filename=filename) as stat_dict_filepath:
            processor = MeanVarianceSpeakerNormalization(field, stats_path=stat_dict_filepath, mask_field=mask_field)
            processor.process(training_example)

        output_tensor = training_example[field]
        torch.testing.assert_close(output_tensor, expected_tensor)



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
# pylint: skip-file

import torch
import torch.nn.functional as F
from megatron.core import InferenceParams
from megatron.core.extensions.transformer_engine import SplitAlongDim
from megatron.core.fusions.fused_bias_geglu import bias_geglu_impl
from megatron.core.fusions.fused_bias_gelu import bias_gelu_impl
from megatron.core.fusions.fused_bias_swiglu import bias_swiglu_impl
from megatron.core.models.common.embeddings.language_model_embedding import LanguageModelEmbedding
from megatron.core.models.common.embeddings.rotary_pos_embedding import apply_rotary_pos_emb
from megatron.core.packed_seq_params import PackedSeqParams
from megatron.core.transformer.attention import SelfAttention
from megatron.core.transformer.mlp import MLP
from megatron.core.transformer.moe.experts import SequentialMLP
from megatron.core.transformer.transformer_block import TransformerBlock
from megatron.core.transformer.transformer_layer import TransformerLayer
from megatron.core.utils import make_viewless_tensor
from torch import Tensor

from nemo.collections.common.modules.adapters.parallel_adapters import (
    AdapterName,
    InfusedAdapterConfig,
    Lora4HtoHAdapterConfig,
    LoraDenseAttentionAdapterConfig,
    LoraHto4HAdapterConfig,
    LoraKQVAdapterConfig,
    LoraMoe4HtoHAdapterConfig,
    LoraMoeHto4HAdapterConfig,
    LoraUnfusedHto4HAdapterConfig,
    LoraUnfusedKQVAdapterConfig,
    MLPHeadAdapterConfig,
    MLPInfusedAdapterConfig,
    ParallelLinearAdapterConfig,
    PromptEncoderAdapterConfig,
)
from nemo.core import adapter_mixins


def swap_mcore_mixin(module, mcore_mixin):
    """
    Casts module to mcore_mixin and register corresponding adapters.
    """
    module.__class__ = mcore_mixin
    module.mcore_register_adapters()


class MCoreAdapterModuleMixin(adapter_mixins.AdapterModuleMixin):
    def mcore_register_adapters(self):
        """
        Performs any necessary setup after swapping class.
        Must use self.set_accepted_adapter_types([<NeMo adapter config>_target_]) to register adapter.
        """
        raise NotImplementedError("Mcore mixins should implement setup_adapters on a subclass of MyBase")


class MCoreTransformerBlockMixin(TransformerBlock, MCoreAdapterModuleMixin):
    def mcore_register_adapters(self):
        """
        Setup NeMo (canonical) Adapter to this MCore layer.
        """
        self.set_accepted_adapter_types([MLPHeadAdapterConfig._target_])

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: Tensor,
        context: Tensor = None,
        context_mask: Tensor = None,
        rotary_pos_emb: Tensor = None,
        rotary_pos_cos: Tensor = None,
        rotary_pos_sin: Tensor = None,
        attention_bias: Tensor = None,
        inference_params: InferenceParams = None,
        packed_seq_params: PackedSeqParams = None,
        sequence_len_offset=None,
    ):
        hidden_states = super().forward(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            context=context,
            context_mask=context_mask,
            rotary_pos_emb=rotary_pos_emb,
            rotary_pos_cos=rotary_pos_cos,
            rotary_pos_sin=rotary_pos_sin,
            inference_params=inference_params,
            packed_seq_params=packed_seq_params,
        )

        mlp_head_adapter = self.get_adapter_module(AdapterName.MLP_HEAD_ADAPTER)
        if mlp_head_adapter and self.adapter_cfg[AdapterName.MLP_HEAD_ADAPTER]['enabled']:
            hidden_states = mlp_head_adapter(hidden_states)

        return hidden_states


class MCoreSelfAttentionMixin(SelfAttention, MCoreAdapterModuleMixin):
    def mcore_register_adapters(self):
        """
        Setup NeMo LoRA or IA3 adapter to this MCore layer.
        """
        self.set_accepted_adapter_types(
            [
                LoraUnfusedKQVAdapterConfig._target_,
                LoraKQVAdapterConfig._target_,
                LoraDenseAttentionAdapterConfig._target_,
                InfusedAdapterConfig._target_,
            ]
        )
        self.linear_qkv.return_layernorm_output = True  # need layernorm output for lora mlp
        if (
            self.config.sequence_parallel
            and hasattr(self.linear_qkv, "return_layernorm_output_gathered")
            and not self.linear_qkv.ub_overlap_ag
        ):
            # for LoRA SP, TE v1.5 can return layernorm output gathered so there is no need
            # to perform the redundant gather in the adapter module, unless TP communication
            # overlap is used.
            self.linear_qkv.return_layernorm_output_gathered = True

    def get_query_key_value_tensors(self, hidden_states, key_value_states=None):
        """
        Derives `query`, `key` and `value` tensors from `hidden_states`.
        """
        # Attention heads [sq, b, h] --> [sq, b, ng * (np/ng + 2) * hn)]
        linear_qkv_output, _ = self.linear_qkv(hidden_states)
        layernorm_output = None

        # In megatron/core/models/gpt/gpt_layer_specs.py when fused module is used(e.g. TELayerNormColumnParallelLinear)
        # both LN and qkv will be returned.
        # In nemo/collections/nlp/models/language_modeling/megatron/falcon/falcon_spec.py TEColumnParallelLinear(non-fused) is used for linear_qkv,
        # which only returns linear.
        if isinstance(linear_qkv_output, tuple):
            if len(linear_qkv_output) == 2:  # fused module, qkv&LN
                mixed_qkv, layernorm_output = linear_qkv_output
            else:
                raise ValueError(f"Unexpected number of outputs from linear_qkv output: {len(linear_qkv_output)}")
        else:  # for qkv&LN not fused only mixed_qkv
            mixed_qkv = linear_qkv_output

        # LoRA logic
        if self.is_adapter_available():
            lora_kqv_adapter = self.get_adapter_module(AdapterName.LORA_KQV_ADAPTER)
            if lora_kqv_adapter and self.adapter_cfg[AdapterName.LORA_KQV_ADAPTER]['enabled']:
                if layernorm_output is not None:
                    lora_mixed_qkv = lora_kqv_adapter(layernorm_output)
                else:
                    lora_mixed_qkv = lora_kqv_adapter(hidden_states)

                mixed_qkv = mixed_qkv + lora_mixed_qkv

        # [sq, b, hp] --> [sq, b, ng, (np/ng + 2) * hn]
        new_tensor_shape = mixed_qkv.size()[:-1] + (
            self.num_query_groups_per_partition,
            (
                (self.num_attention_heads_per_partition // self.num_query_groups_per_partition + 2)
                * self.hidden_size_per_attention_head
            ),
        )
        mixed_qkv = mixed_qkv.view(*new_tensor_shape)

        # [sq, b, ng, (np/ng + 2) * hn] --> [sq, b, ng, np/ng * hn], [sq, b, ng, hn], [sq, b, ng, hn]
        split_arg_list = [
            (
                self.num_attention_heads_per_partition
                // self.num_query_groups_per_partition
                * self.hidden_size_per_attention_head
            ),
            self.hidden_size_per_attention_head,
            self.hidden_size_per_attention_head,
        ]

        if SplitAlongDim is not None:

            # [sq, b, ng, (np/ng + 2) * hn] --> [sq, b, ng, np/ng * hn], [sq, b, ng, hn], [sq, b, ng, hn]
            (query, key, value) = SplitAlongDim(
                mixed_qkv,
                3,
                split_arg_list,
            )
        else:

            # [sq, b, ng, (np/ng + 2) * hn] --> [sq, b, ng, np/ng * hn], [sq, b, ng, hn], [sq, b, ng, hn]
            (query, key, value) = torch.split(
                mixed_qkv,
                split_arg_list,
                dim=3,
            )

        # [sq, b, ng, np/ng * hn] -> [sq, b, np, hn]
        query = query.reshape(query.size(0), query.size(1), -1, self.hidden_size_per_attention_head)

        if self.is_adapter_available():
            key_infused_adapter = self.get_adapter_module(AdapterName.KEY_INFUSED)
            value_infused_adapter = self.get_adapter_module(AdapterName.VALUE_INFUSED)
            if key_infused_adapter and self.adapter_cfg[AdapterName.KEY_INFUSED]['enabled']:
                assert value_infused_adapter is not None, "Expected value_infused_adapter not found!"
                kls = key.shape
                key = key_infused_adapter(key.reshape(kls[0], kls[1], -1)).reshape(kls).to(query.dtype)
            if value_infused_adapter and self.adapter_cfg[AdapterName.VALUE_INFUSED]['enabled']:
                assert key_infused_adapter is not None, "Expected key_infused_adapter not found!"
                vls = value.shape
                value = value_infused_adapter(value.reshape(vls[0], vls[1], -1)).reshape(vls).to(query.dtype)

            lora_unfused_kqv_adapter = self.get_adapter_module(AdapterName.LORA_UNFUSED_KQV_ADAPTER)
            if lora_unfused_kqv_adapter and self.adapter_cfg[AdapterName.LORA_UNFUSED_KQV_ADAPTER]['enabled']:
                assert lora_kqv_adapter is None
                if layernorm_output is not None:
                    lq, lk, lv = lora_unfused_kqv_adapter(layernorm_output)
                else:
                    lq, lk, lv = lora_unfused_kqv_adapter(hidden_states)
                query, key, value = query + lq, key + lk, value + lv
        return query, key, value

    def forward(
        self,
        hidden_states,
        attention_mask,
        key_value_states=None,
        inference_params=None,
        rotary_pos_emb=None,
        packed_seq_params=None,
        rotary_pos_cos=None,
        rotary_pos_sin=None,
        attention_bias=None,
        sequence_len_offset=None,
    ):
        # hidden_states: [sq, b, h]

        # For self attention we just duplicate the rotary_pos_emb if it isn't already
        if rotary_pos_emb is not None and not isinstance(rotary_pos_emb, tuple):
            rotary_pos_emb = (rotary_pos_emb,) * 2

        # =====================
        # Query, Key, and Value
        # =====================
        # Get the query, key and value tensors based on the type of attention -
        # self or cross attn.
        query, key, value = self.get_query_key_value_tensors(hidden_states, key_value_states)

        # ===================================================
        # Adjust key, value, and rotary_pos_emb for inference
        # ===================================================
        query, key, value, rotary_pos_emb, attn_mask_type = self._adjust_key_value_for_inference(
            inference_params, query, key, value, rotary_pos_emb
        )

        if packed_seq_params is not None:
            query = query.squeeze(1)
            key = key.squeeze(1)
            value = value.squeeze(1)

        # ================================================
        # relative positional embedding (rotary embedding)
        # ================================================
        if rotary_pos_emb is not None:
            q_pos_emb, k_pos_emb = rotary_pos_emb

            if packed_seq_params is not None:
                if packed_seq_params.cu_seqlens_q_padded is not None:
                    cu_seqlens_q = packed_seq_params.cu_seqlens_q_padded
                else:
                    cu_seqlens_q = packed_seq_params.cu_seqlens_q
                if packed_seq_params.cu_seqlens_kv_padded is not None:
                    cu_seqlens_kv = packed_seq_params.cu_seqlens_kv_padded
                else:
                    cu_seqlens_kv = packed_seq_params.cu_seqlens_kv
            else:
                cu_seqlens_q = cu_seqlens_kv = None

            query = apply_rotary_pos_emb(query, q_pos_emb, config=self.config, cu_seqlens=cu_seqlens_q)
            key = apply_rotary_pos_emb(key, k_pos_emb, config=self.config, cu_seqlens=cu_seqlens_kv)
            # TODO, can apply positional embedding to value_layer so it has
            # absolute positional embedding.
            # otherwise, only relative positional embedding takes effect
            # value_layer = apply_rotary_pos_emb(value_layer, k_pos_emb)

        # ==================================
        # core attention computation
        # ==================================

        if self.checkpoint_core_attention:
            core_attn_out = self._checkpointed_attention_forward(
                query,
                key,
                value,
                attention_mask,
                attn_mask_type=attn_mask_type,
                packed_seq_params=packed_seq_params,
            )
        else:
            core_attn_out = self.core_attention(
                query,
                key,
                value,
                attention_mask,
                attn_mask_type=attn_mask_type,
                packed_seq_params=packed_seq_params,
            )

        if packed_seq_params is not None:
            # reshape to same output shape as unpacked case
            # (t, np, hn) -> (t, b=1, h=np*hn)
            # t is the pack size = sum (sq_i)
            # note that batch is a dummy dimension in the packed case
            core_attn_out = core_attn_out.reshape(core_attn_out.size(0), 1, -1)

        # =================
        # Output. [sq, b, h]
        # =================

        output, bias = self.linear_proj(core_attn_out)
        # LoRA logic
        if self.is_adapter_available():
            lora_linear_proj_adapter = self.get_adapter_module(AdapterName.LORA_DENSE_ATTENTION_ADAPTER)
            if lora_linear_proj_adapter and self.adapter_cfg[AdapterName.LORA_DENSE_ATTENTION_ADAPTER]['enabled']:
                lora_output = lora_linear_proj_adapter(core_attn_out)
                output = output + lora_output

        return output, bias


class MCoreMLPMixin(MLP, MCoreAdapterModuleMixin):
    def mcore_register_adapters(self):
        """
        Setup NeMo IA3 and LoRA adapter to this MCore layer.
        """
        self.set_accepted_adapter_types(
            [
                LoraUnfusedHto4HAdapterConfig._target_,
                LoraHto4HAdapterConfig._target_,
                Lora4HtoHAdapterConfig._target_,
                LoraMoeHto4HAdapterConfig._target_,
                LoraMoe4HtoHAdapterConfig._target_,
                MLPInfusedAdapterConfig._target_,
            ]
        )  # only self attn (packed qkv) for now
        self.linear_fc1.return_layernorm_output = True  # need layernorm output for lora mlp
        if (
            self.config.sequence_parallel
            and hasattr(self.linear_fc1, "return_layernorm_output_gathered")
            and not self.config.tp_comm_overlap
        ):
            # for LoRA SP, TE v1.5 can return layernorm output gathered so there is no need
            # to perform the redundant gather in the adapter module, unless TP communication
            # overlap is used.
            self.linear_fc1.return_layernorm_output_gathered = True

    def forward(self, hidden_states, expert_idx=None):
        # [s, b, 4 * h/p]
        output = self.linear_fc1(hidden_states)
        if isinstance(output, tuple) and len(output) == 2:
            intermediate_parallel, bias_parallel = output
            if isinstance(intermediate_parallel, tuple) and len(intermediate_parallel) == 2:
                intermediate_parallel, layernorm_output = intermediate_parallel
            else:
                layernorm_output = hidden_states
        else:
            # self.linear_fc1.te_return_bias == True
            intermediate_parallel, bias_parallel, layernorm_output = output

        # LoRA logic
        if self.is_adapter_available():
            lora_adapter = None
            lora_fc1_adapter = self.get_adapter_module(AdapterName.LORA_Hto4H_ADAPTER)
            lora_unfused_fc1_adapter = self.get_adapter_module(AdapterName.LORA_UNFUSED_Hto4H_ADAPTER)
            lora_moe_fc1_adapter = self.get_adapter_module(AdapterName.LORA_MOE_Hto4H_ADAPTER)
            if lora_fc1_adapter and self.adapter_cfg[AdapterName.LORA_Hto4H_ADAPTER]['enabled']:
                lora_adapter = lora_fc1_adapter
            if lora_unfused_fc1_adapter and self.adapter_cfg[AdapterName.LORA_UNFUSED_Hto4H_ADAPTER]['enabled']:
                assert lora_adapter is None, "Expected only one of LORA_Hto4H_ADAPTER or LORA_UNFUSED_Hto4H_ADAPTER"
                lora_adapter = lora_unfused_fc1_adapter

            lora_output = 0
            if lora_adapter:
                lora_output = lora_adapter(layernorm_output)
            elif lora_moe_fc1_adapter and self.adapter_cfg[AdapterName.LORA_MOE_Hto4H_ADAPTER]['enabled']:
                lora_output = lora_moe_fc1_adapter(layernorm_output, expert_idx)
            intermediate_parallel = intermediate_parallel + lora_output

        if self.config.bias_activation_fusion:
            if self.activation_func == F.gelu:
                if self.config.gated_linear_unit:
                    intermediate_parallel = bias_geglu_impl(intermediate_parallel, bias_parallel)
                else:
                    assert self.config.add_bias_linear is True
                    intermediate_parallel = bias_gelu_impl(intermediate_parallel, bias_parallel)
            elif self.activation_func == F.silu and self.config.gated_linear_unit:
                intermediate_parallel = bias_swiglu_impl(
                    intermediate_parallel,
                    bias_parallel,
                    self.config.activation_func_fp8_input_store,
                )

            else:
                raise ValueError("Only support fusion of gelu and swiglu")
        else:
            if bias_parallel is not None:
                intermediate_parallel = intermediate_parallel + bias_parallel
            if self.config.gated_linear_unit:

                def glu(x):
                    x = torch.chunk(x, 2, dim=-1)
                    return self.config.activation_func(x[0]) * x[1]

                intermediate_parallel = glu(intermediate_parallel)
            else:
                intermediate_parallel = self.activation_func(intermediate_parallel)

        infused_adapter = self.get_adapter_module(AdapterName.MLP_INFUSED)
        if infused_adapter:
            intermediate_parallel = infused_adapter(intermediate_parallel)

        # [s, b, h]
        output, output_bias = self.linear_fc2(intermediate_parallel)

        # LoRA logic
        if self.is_adapter_available():
            lora_fc2_adapter = self.get_adapter_module(AdapterName.LORA_4HtoH_ADAPTER)
            lora_moe_fc2_adapter = self.get_adapter_module(AdapterName.LORA_MOE_4HtoH_ADAPTER)

            lora_output = 0
            if lora_fc2_adapter and self.adapter_cfg[AdapterName.LORA_4HtoH_ADAPTER]['enabled']:
                lora_output = lora_fc2_adapter(intermediate_parallel)
            elif lora_moe_fc2_adapter and self.adapter_cfg[AdapterName.LORA_MOE_4HtoH_ADAPTER]['enabled']:
                lora_output = lora_moe_fc2_adapter(intermediate_parallel, expert_idx)

            output = output + lora_output

        return output, output_bias


class MCoreSequentialMLPMixin(SequentialMLP, MCoreAdapterModuleMixin):
    def mcore_register_adapters(self):
        """
        We don't want the SequentialMLP layer to take any adapters. We only want to override the forward() behavior
        """
        pass

    def forward(self, permuted_local_hidden_states, tokens_per_expert):
        output_local = torch.zeros_like(permuted_local_hidden_states)
        output_bias_local = None
        if self.add_bias:
            output_bias_local = torch.zeros_like(permuted_local_hidden_states)

        cumsum_num_tokens = torch.cumsum(tokens_per_expert, dim=0)
        # Insert zero at the begining for offset index's convenience
        zero_tensor = torch.zeros(1, dtype=torch.long, device=cumsum_num_tokens.device)
        cumsum_num_tokens = torch.cat((zero_tensor, cumsum_num_tokens))
        for expert_num, expert in enumerate(self.local_experts):
            start = cumsum_num_tokens[expert_num]
            end = cumsum_num_tokens[expert_num + 1]
            hidden = permuted_local_hidden_states[start:end]
            output, output_bias = expert(hidden, expert_num)  # expert: MLP

            output_local[start:end] = output
            if self.add_bias:
                output_bias = output_bias.expand_as(output)
                output_bias_local[start:end, :] = output_bias

        return output_local, output_bias_local


class MCoreGPTEmbeddingMixin(LanguageModelEmbedding, MCoreAdapterModuleMixin):
    def mcore_register_adapters(self):
        """
        Setup NeMo ptuning adapter to this MCore layer.
        """
        self.set_accepted_adapter_types([PromptEncoderAdapterConfig._target_])

    def forward(self, input_ids, position_ids):
        encoder_input = super().forward(input_ids, position_ids)

        if self.is_adapter_available():
            _sq, _bs, _hs = encoder_input.size()
            ptuning_adapter = self.get_adapter_module(AdapterName.PTUNING_ADAPTER)
            v = ptuning_adapter.virtual_tokens
            if (
                ptuning_adapter and self.adapter_cfg[AdapterName.PTUNING_ADAPTER]['enabled'] and _sq >= v
            ):  # The sequence should be longer the v to insert virtual embeddings.
                virtual_embeddings = ptuning_adapter(_bs)
                encoder_input = encoder_input[
                    v:, :, :
                ]  # the first v tokens are pads so that they can be swapped out with virtual embeddings.
                encoder_input = torch.concat([virtual_embeddings, encoder_input], dim=0)
        return encoder_input


class MCoreTransformerLayerMixin(TransformerLayer, MCoreAdapterModuleMixin):
    def mcore_register_adapters(self):
        """
        Setup NeMo (canonical) Adapter to this MCore layer.
        """
        self.set_accepted_adapter_types([ParallelLinearAdapterConfig._target_])

    def forward(
        self,
        hidden_states,
        attention_mask,
        context=None,
        context_mask=None,
        rotary_pos_emb=None,
        inference_params=None,
        packed_seq_params=None,
    ):
        # hidden_states: [s, b, h]

        # Residual connection.
        residual = hidden_states

        # Optional Input Layer norm
        input_layernorm_output = self.input_layernorm(hidden_states)

        # Self attention.
        attention_output_with_bias = self.self_attention(
            input_layernorm_output,
            attention_mask=attention_mask,
            inference_params=inference_params,
            rotary_pos_emb=rotary_pos_emb,
            packed_seq_params=packed_seq_params,
        )

        # adapter logic
        if self.is_adapter_available():
            adapter_1 = self.get_adapter_module(AdapterName.PRE_ATTN_ADAPTER)
            if adapter_1 and self.adapter_cfg[AdapterName.PRE_ATTN_ADAPTER]['enabled']:
                attention_output, bias = attention_output_with_bias
                attention_output = (
                    adapter_1(attention_output) + attention_output
                )  # simple adapter call with residual connection
                attention_output_with_bias = (attention_output, bias)

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.self_attn_bda(self.training, self.config.bias_dropout_fusion)(
                attention_output_with_bias, residual, self.hidden_dropout
            )

        # Residual connection.
        residual = hidden_states

        # Optional Layer norm after self-attention
        pre_cross_attn_layernorm_output = self.pre_cross_attn_layernorm(hidden_states)

        # Cross attention.
        attention_output_with_bias = self.cross_attention(
            pre_cross_attn_layernorm_output,
            attention_mask=context_mask,
            key_value_states=context,
            inference_params=inference_params,
        )

        if isinstance(attention_output_with_bias, dict) and "context" in attention_output_with_bias:
            context = attention_output_with_bias["context"]

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.cross_attn_bda(self.training, self.config.bias_dropout_fusion)(
                attention_output_with_bias, residual, self.hidden_dropout
            )

        # Residual connection.
        residual = hidden_states

        # Optional Layer norm post the cross-attention.
        pre_mlp_layernorm_output = self.pre_mlp_layernorm(hidden_states)

        # MLP.
        mlp_output_with_bias = self.mlp(pre_mlp_layernorm_output)

        # adapter logic
        if self.is_adapter_available():
            adapter_2 = self.get_adapter_module(AdapterName.POST_ATTN_ADAPTER)
            if adapter_2 and self.adapter_cfg[AdapterName.POST_ATTN_ADAPTER]['enabled']:
                mlp_output, bias = mlp_output_with_bias
                mlp_output = adapter_2(mlp_output) + mlp_output  # simple adapter call with residual connection
                mlp_output_with_bias = (mlp_output, bias)

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.mlp_bda(self.training, self.config.bias_dropout_fusion)(
                mlp_output_with_bias, residual, self.hidden_dropout
            )

        # Jit compiled function creates 'view' tensor. This tensor
        # potentially gets saved in the MPU checkpoint function context,
        # which rejects view tensors. While making a viewless tensor here
        # won't result in memory savings (like the data loader, or
        # p2p_communication), it serves to document the origin of this
        # 'view' tensor.
        output = make_viewless_tensor(inp=hidden_states, requires_grad=hidden_states.requires_grad, keep_graph=True)

        return output, context



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import numpy as np
import torch
from kaldialign import align
from omegaconf import open_dict

from nemo.collections.asr.models import ASRModel, EncDecRNNTModel
from nemo.collections.asr.parts.utils.confidence_metrics import (
    auc_nt,
    auc_pr,
    auc_roc,
    auc_yc,
    ece,
    nce,
    save_confidence_hist,
    save_custom_confidence_curve,
    save_nt_curve,
    save_pr_curve,
    save_roc_curve,
)
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis


def get_correct_marks(r: Union[List[int], List[str]], h: Union[List[int], List[str]]) -> List[bool]:
    """Get correct marks by aligning the reference text with a hypothesis.

    This method considers only insertions and substitutions as incorrect marks.
    """
    alignment = align(
        [str(rr) for rr in r],
        [str(hh) for hh in h],
        "<eps>",
    )
    return [a == b for a, b in alignment if b != "<eps>"]


def get_token_targets_with_confidence(hyp: Hypothesis) -> List[Tuple[str, float]]:
    return [(y, c) for y, c in zip(hyp.y_sequence, hyp.token_confidence)]


def get_word_targets_with_confidence(hyp: Hypothesis) -> List[Tuple[str, float]]:
    return [(y, c) for y, c in zip(hyp.words, hyp.word_confidence)]


def run_confidence_benchmark(
    model: ASRModel,
    target_level: str,
    filepaths: List[str],
    reference_texts: List[str],
    batch_size: int = 8,
    num_workers: int = 4,
    plot_dir: Optional[Union[str, Path]] = None,
    use_amp: Optional[bool] = False,
):
    """Run benchmark and plot histograms and curves, if plot_dir is provided.

    Returns:
        Dictionary with benchmark results of the following scheme:
        `level: (auc_roc, auc_pr, auc_nt, nce, ece, auc_yc, std_yc, max_yc)` with `level` being 'token' or 'word'.
    """
    draw_plot = plot_dir is not None
    if isinstance(plot_dir, str):
        plot_dir = Path(plot_dir)

    # transcribe audio
    with torch.amp.autocast(model.device.type, enabled=use_amp):
        with torch.no_grad():
            transcriptions = model.transcribe(
                audio=filepaths, batch_size=batch_size, return_hypotheses=True, num_workers=num_workers
            )

    levels = []
    if target_level != "word":
        levels.append("token")
    if target_level != "token":
        levels.append("word")
    results = {}
    for level in levels:
        if level == "token":
            targets_with_confidence = [get_token_targets_with_confidence(tran) for tran in transcriptions]
            correct_marks = [
                get_correct_marks(model.tokenizer.text_to_ids(r), model.tokenizer.text_to_ids(h.text))
                for r, h in zip(reference_texts, transcriptions)
            ]
        else:  # "word"
            targets_with_confidence = [get_word_targets_with_confidence(tran) for tran in transcriptions]
            correct_marks = [get_correct_marks(r.split(), h.words) for r, h in zip(reference_texts, transcriptions)]

        y_true, y_score = np.array(
            [[f, p[1]] for cm, twc in zip(correct_marks, targets_with_confidence) for f, p in zip(cm, twc)]
        ).T
        # output scheme: yc.mean(), yc.max(), yc.std() or yc.mean(), yc.max(), yc.std(), (thresholds, yc)
        result_yc = auc_yc(y_true, y_score, return_std_maximum=True, return_curve=draw_plot)
        # output scheme: ece or ece, (thresholds, ece_curve)
        results_ece = ece(y_true, y_score, return_curve=draw_plot)
        results[level] = [
            auc_roc(y_true, y_score),
            auc_pr(y_true, y_score),
            auc_nt(y_true, y_score),
            nce(y_true, y_score),
            results_ece if isinstance(results_ece, float) else results_ece[0],
        ] + list(result_yc[:3])

        if draw_plot:
            os.makedirs(plot_dir, exist_ok=True)

            mask_correct = y_true == 1
            y_score_correct = y_score[mask_correct]
            y_score_incorrect = y_score[~mask_correct]
            # histogram of the correct distribution
            save_confidence_hist(y_score_correct, plot_dir, level + "_" + "hist_correct")
            # histogram of the incorrect distribution
            save_confidence_hist(y_score_incorrect, plot_dir, level + "_" + "hist_incorrect")
            # AUC-ROC curve
            save_roc_curve(y_true, y_score, plot_dir, level + "_" + "roc")
            # AUC-PR curve
            save_pr_curve(y_true, y_score, plot_dir, level + "_" + "pr")
            # AUC-NT curve
            save_nt_curve(y_true, y_score, plot_dir, level + "_" + "nt")
            # AUC-YC curve
            yc_thresholds, yc_values = result_yc[-1]
            save_custom_confidence_curve(
                yc_thresholds,
                yc_values,
                plot_dir,
                level + "_" + "yc",
                "Threshold",
                "True positive rate − False Positive Rate",
            )
            # ECE curve
            ece_thresholds, ece_values = results_ece[-1]
            ece_values /= max(ece_values)
            save_custom_confidence_curve(
                ece_thresholds, ece_values, plot_dir, level + "_" + "ece", "Threshold", "|Accuracy − Confidence score|"
            )

    return results


def apply_confidence_parameters(decoding_cfg, hp):
    """Apply parameters from a parameter grid to a decoding config.

    Returns:
        Updated decoding config.
    """
    new_decoding_cfg = copy.deepcopy(decoding_cfg)
    confidence_cfg_fields = ("aggregation", "exclude_blank", "tdt_include_duration")
    confidence_method_cfg_fields = ("name", "alpha", "entropy_type", "entropy_norm")
    with open_dict(new_decoding_cfg):
        for p, v in hp.items():
            if p in confidence_cfg_fields:
                new_decoding_cfg.confidence_cfg[p] = v
            elif p in confidence_method_cfg_fields:
                new_decoding_cfg.confidence_cfg.method_cfg[p] = v
    return new_decoding_cfg
